# -*- coding: utf-8 -*-


class InvalidElementValueError(ValueError):
    def __init__(self, elem_name, elem_cls, value):
        self.elem_name = elem_name
        self.elem_cls = elem_cls
        self.value = value

    def __str__(self):
        return ("Could not assign value '{value}' to element '{elem_name}': "
                "element requires attributes or value is not instance of '{elem_cls}'. "
                "For attributes modification use properties: element.attribute_name = attribute_value. "
                "For multiple allowed elements use list: category_element = ['cat1', 'cat2', 'cat3']"
                .format(value=self.value, elem_name=self.elem_name, elem_cls=self.elem_cls))


class InvalidRssItemError(ValueError):
    pass


class InvalidRssItemAttributesError(ValueError):
    def __init__(self, rss_element, required_attrs, content_arg):
        self.rss_element = rss_element
        self.required_attrs = required_attrs
        self.content_arg = content_arg

    def __str__(self):
        if self.content_arg:
            return "The next required attributes of RSS element '{}' ({}) "\
                   "or required content ('{}' argument) are not set" \
                   .format(self.rss_element, ", ".join(str(a) for a in self.required_attrs), self.content_arg)
        return "The next required attributes of RSS element '{}' are not set: {}" \
            .format(self.rss_element, ", ".join(str(a) for a in self.required_attrs))


class NoNamespaceURIError(ValueError):
    pass



# -*- coding: utf-8 -*-

import re
from .nscomponent import BaseNSComponent


class ItemElementAttribute(BaseNSComponent):
    def __init__(self, value=None, serializer=lambda x: str(x),
                 required=False, is_content=False, **kwargs):
        """
        Construct attribute of an element

        Parameters
        ----------
        value : Any
            Default attribute value, None means unassigned
        serializer : callable
            Function to serialize the attribute value to string for a XML document
        required : bool
             Whether the attribute is required
        is_content : bool
             Whether the "attribute" is an element content
        """
        super(ItemElementAttribute, self).__init__(**kwargs)
        if is_content and self.ns_uri:
            raise ValueError("Content cannot have namespace")
        self.__required = required
        self.__is_content = is_content
        self.serializer = serializer
        self.value = value

    @property
    def required(self):
        """
        Returns
        -------
        bool
            Whether the attribute is required
        """
        return self.__required

    @property
    def is_content(self):
        """
        Returns
        -------
        bool
            Whether the attribute is an element content
        """
        return self.__is_content

    def __repr__(self):
        s_match = re.match(r'^[^(]+\((.*?)\)$', super(ItemElementAttribute, self).__repr__())
        s_repr = ", " + s_match.group(1) if s_match else ''
        return "{}(value={!r}, serializer={!r}, required={!r}, is_content={!r}{})"\
            .format(self.__class__.__name__, self.value, self.serializer,
                    self.__required, self.__is_content, s_repr)



# -*- coding: utf-8 -*-

from collections import Counter
import itertools

from unittest.util import safe_repr
import difflib
import re
from os.path import commonprefix
import sys
import pprint
import six

from twisted.python.failure import Failure
from scrapy.pipelines import ItemPipelineManager
from lxml import etree
from xmlunittest import XmlTestCase

from scrapy_rss.meta import ItemElement, MultipleElements
from scrapy_rss.items import RssItem

try:
    from collections.abc import Mapping
except ImportError:
    from collections import Mapping

try:
    from unittest.util import _common_shorten_repr, _shorten
except ImportError:
    _MAX_LENGTH = 80
    _PLACEHOLDER_LEN = 12
    _MIN_BEGIN_LEN = 5
    _MIN_END_LEN = 5
    _MIN_COMMON_LEN = 5
    _MIN_DIFF_LEN = _MAX_LENGTH - \
                    (_MIN_BEGIN_LEN + _PLACEHOLDER_LEN + _MIN_COMMON_LEN +
                     _PLACEHOLDER_LEN + _MIN_END_LEN)
    assert _MIN_DIFF_LEN >= 0


    def _shorten(s, prefixlen, suffixlen):
        skip = len(s) - prefixlen - suffixlen
        if skip > _PLACEHOLDER_LEN:
            s = '%s[%d chars]%s' % (s[:prefixlen], skip, s[len(s) - suffixlen:])
        return s


    def _common_shorten_repr(*args):
        args = tuple(map(safe_repr, args))
        maxlen = max(map(len, args))
        if maxlen <= _MAX_LENGTH:
            return args

        prefix = commonprefix(args)
        prefixlen = len(prefix)

        common_len = _MAX_LENGTH - \
                     (maxlen - prefixlen + _MIN_BEGIN_LEN + _PLACEHOLDER_LEN)
        if common_len > _MIN_COMMON_LEN:
            assert _MIN_BEGIN_LEN + _PLACEHOLDER_LEN + _MIN_COMMON_LEN + \
                   (maxlen - prefixlen) < _MAX_LENGTH
            prefix = _shorten(prefix, _MIN_BEGIN_LEN, common_len)
            return tuple(prefix + s[prefixlen:] for s in args)

        prefix = _shorten(prefix, _MIN_BEGIN_LEN, _MIN_COMMON_LEN)
        return tuple(prefix + _shorten(s[prefixlen:], _MIN_DIFF_LEN, _MIN_END_LEN)
                     for s in args)


def get_dict_attr(obj,attr):
    for obj in [obj]+obj.__class__.mro():
        if attr in obj.__dict__:
            return obj.__dict__[attr]
    raise AttributeError


iteritems = getattr(dict, 'iteritems', dict.items) # py2-3 compatibility


class FrozenDict(Mapping):
    """
    A simple immutable wrapper around dictionaries.
    It can be used as a drop-in replacement for dictionaries where immutability is desired.
    """

    dict_cls = dict

    def __init__(self, *args, **kwargs):
        self._dict = self.dict_cls(*args, **kwargs)
        self._hash = None

    def __getitem__(self, key):
        return self._dict[key]

    def __contains__(self, key):
        return key in self._dict

    def copy(self, **add_or_replace):
        return self.__class__(self, **add_or_replace)

    def __iter__(self):
        return iter(self._dict)

    def __len__(self):
        return len(self._dict)

    def __repr__(self):
        return '<%s %r>' % (self.__class__.__name__, self._dict)

    def __hash__(self):
        if self._hash is None:
            h = 0
            for key, value in iteritems(self._dict):
                h ^= hash((key, value))
            self._hash = h
        return self._hash


class RaisedItemPipelineManager(ItemPipelineManager):
    def process_item(self, item, spider):
        d = super(RaisedItemPipelineManager, self).process_item(item, spider)
        if isinstance(d.result, Failure):
            failure = d.result
            d.addErrback(lambda failure: None)  # workaround for Python 2.*
            print(failure.getTraceback())
            failure.raiseException()
        return d


class UnorderedXmlTestCase(XmlTestCase):
    """
    Expand XmlTestCase functionality with unordered XML equivalence testing.
    """

    @classmethod
    def _xml_to_tuple(cls, element):
        return (element.tag,
                FrozenDict(element.nsmap),
                FrozenDict(element.attrib),
                FrozenDict(Counter(t for t in element.itertext() if t.strip())),
                frozenset(cls._xml_to_tuple(child) for child in element.getchildren()))

    @staticmethod
    def _str_to_bytes(data):
        if isinstance(data, str):
            return data.encode(encoding='utf-8')
        if not isinstance(data, bytes):
            raise ValueError("Passing data must be string or bytes array")
        return data

    def _getAssertEqualityFunc(self, first, second):
        for (first_cls, second_cls) in ((first.__class__, second.__class__),
                                        (first.__class__, None),
                                        (None, second.__class__)):
            if (first_cls, second_cls) in self._type_equality_funcs:
                asserter = self._type_equality_funcs[(first_cls, second_cls)]
                if asserter is not None:
                    if isinstance(asserter, str):
                        asserter = getattr(self, asserter)
                    return asserter
        return super(UnorderedXmlTestCase, self)._getAssertEqualityFunc(first, second)

    def assertUnorderedXmlEquivalentOutputs(self, data, expected, excepted_elements = ('lastBuildDate', 'generator')):
        """
        Children and text subnodes of each element in XML are considered as unordered set.
        Therefore if two XML files has different order of same elements then these XMLs are same.
        """
        if not excepted_elements:
            excepted_elements = ()
        if isinstance(excepted_elements, six.string_types):
            excepted_elements = (excepted_elements,)

        data = data if isinstance(data, etree._Element) \
            else etree.fromstring(self._str_to_bytes(data))
        for excepted_element in excepted_elements:
            for element in data.xpath('//{}'.format(excepted_element)):
                element.getparent().remove(element)
        data_tuple = self._xml_to_tuple(data)

        expected = expected if isinstance(expected, etree._Element) \
            else etree.fromstring(self._str_to_bytes(expected))
        for excepted_element in excepted_elements:
            for element in expected.xpath('//{}'.format(excepted_element)):
                element.getparent().remove(element)
        expected_tuple = self._xml_to_tuple(expected)

        if data_tuple != expected_tuple:
            self.fail(
                'Feeds are not equivalent accurate within ordering '
                '(taking into consideration excepted nodes {excepted_elements}):\n'
                'Given: {given}\n'
                'Expected: {expected}'
                .format(excepted_elements=excepted_elements,
                        given=etree.tostring(data), expected=etree.tostring(expected)))

    def assertXmlDocument(self, data):
        data = self._str_to_bytes(data)
        return super(UnorderedXmlTestCase, self).assertXmlDocument(data)

    def assertXmlEquivalentOutputs(self, data, expected):
        data = self._str_to_bytes(data)
        expected = self._str_to_bytes(expected)
        return super(UnorderedXmlTestCase, self).assertXmlEquivalentOutputs(data, expected)

    def assertSequenceEqual(self, seq1, seq2, msg=None, seq_type=None):
        """An equality assertion for ordered sequences (like lists and tuples).

        For the purposes of this function, a valid ordered sequence type is one
        which can be indexed, has a length, and has an equality operator.

        Args:
            seq1: The first sequence to compare.
            seq2: The second sequence to compare.
            seq_type: The expected datatype of the sequences, or None if no
                    datatype should be enforced.
            msg: Optional message to use on failure instead of a list of
                    differences.
        """
        if seq_type is not None:
            seq_type_name = seq_type.__name__
            if not isinstance(seq1, seq_type):
                raise self.failureException('First sequence is not a %s: %s'
                                        % (seq_type_name, safe_repr(seq1)))
            if not isinstance(seq2, seq_type):
                raise self.failureException('Second sequence is not a %s: %s'
                                        % (seq_type_name, safe_repr(seq2)))
        else:
            seq_type_name = "sequence"

        differing = None
        try:
            len1 = len(seq1)
        except (TypeError, NotImplementedError):
            differing = 'First %s has no length.    Non-sequence?' % (
                    seq_type_name)

        if differing is None:
            try:
                len2 = len(seq2)
            except (TypeError, NotImplementedError):
                differing = 'Second %s has no length.    Non-sequence?' % (
                        seq_type_name)

        if differing is None:
            if len1 > len2:
                differing = ('\nFirst %s contains %d additional '
                              'elements.\n' % (seq_type_name, len1 - len2))
                try:
                    differing += ('First extra element %d:\n%s\n' %
                                  (len2, seq1[len2]))
                except (TypeError, IndexError, NotImplementedError):
                    differing += ('Unable to index element %d '
                                  'of first %s\n' % (len2, seq_type_name))
            elif len1 < len2:
                differing = ('\nSecond %s contains %d additional '
                              'elements.\n' % (seq_type_name, len2 - len1))
                try:
                    differing += ('First extra element %d:\n%s\n' %
                                  (len1, seq2[len1]))
                except (TypeError, IndexError, NotImplementedError):
                    differing += ('Unable to index element %d '
                                  'of second %s\n' % (len1, seq_type_name))

            tmp_differing = ''
            for ind, (item1, item2) in enumerate(zip(seq1, seq2)):
                try:
                        self.assertEqual(item1, item2)
                except AssertionError as e:
                    tmp_differing += ('\nFirst differing element %d: %s\n' %
                                      (ind, e.args[0]))
                    break
            if tmp_differing:
                differing = '%ss differ: %s != %s\n%s' % (
                    (seq_type_name.capitalize(),) +
                    _common_shorten_repr(seq1, seq2) + (tmp_differing,))
            else:
                return

        standardMsg = differing
        diffMsg = '\n' + '\n'.join(
            difflib.ndiff(pprint.pformat(seq1).splitlines(),
                          pprint.pformat(seq2).splitlines()))

        standardMsg = self._truncateMessage(standardMsg, diffMsg)
        msg = self._formatMessage(msg, standardMsg)
        self.fail(msg)


class RssTestCase(UnorderedXmlTestCase):
    def __init__(self, *args, **kwargs):
        super(RssTestCase, self).__init__(*args, **kwargs)
        for elem in RssItem().elements.values():
            if isinstance(elem, MultipleElements):
                self.addTypeEqualityFunc((elem.__class__, None), self.assertMultipleRssElementsEqualsToValues)
            else:
                self.addTypeEqualityFunc((elem.__class__, None), self.assertRssElementEqualsToValue)

    def assertRssElementEqualsToValue(self, element, value, msg=None):
        if isinstance(value, ItemElement):
            raise NotImplemented
        self.assertEqual(getattr(element, str(element.content_arg)), value, msg)

    def assertMultipleRssElementsEqualsToValues(self, multiple_element, values, msg=None):
        if isinstance(values, ItemElement):
            raise NotImplemented
        if len(multiple_element) == 1:
            self.assertRssElementEqualsToValue(multiple_element, values, msg)
        else:
            self.assertSequenceEqual([getattr(elem, str(elem.content_arg)) for elem in multiple_element], values, msg)





# -*- coding: utf-8 -*-

import unittest
from parameterized import parameterized
from itertools import chain, product, combinations, permutations
import re
from copy import deepcopy
import six
import sys

from tests.utils import RssTestCase, get_dict_attr

from scrapy_rss.items import RssItem
from scrapy_rss.elements import *
from scrapy_rss.meta import ItemElementAttribute, ItemElement, MultipleElements
from scrapy_rss.exceptions import InvalidElementValueError


ns_attrs = {'attr1': ItemElementAttribute(ns_prefix='prefix1', ns_uri='id1'),
            'prefix2__attr2': ItemElementAttribute(ns_uri='id2'),
            'prefix3__attr3': ItemElementAttribute(ns_prefix='prefix3', ns_uri='id3'),
            'pseudo_prefix4__attr4': ItemElementAttribute(ns_prefix='prefix4', ns_uri='id4')}

ns_elem_names = {'elem1': {'ns_prefix': 'el_prefix1', 'ns_uri': 'el_id1'},
                 'el_prefix2__elem2': {'ns_uri': 'el_id2'},
                 'el_prefix3__elem3': {'ns_prefix': 'el_prefix3', 'ns_uri': 'el_id3'},
                 'el_pseudo_prefix4__elem4': {'ns_prefix': 'el_prefix4', 'ns_uri': 'el_id4'},}

values = [None, 0, 1, '', '1', 'long текст']


class TestRepr(RssTestCase):
    @parameterized.expand((value, required, is_content, ns_prefix, ns_uri)
                          for value in values
                          for required in (True, False)
                          for is_content in (True, False)
                          for ns_prefix in (None, '', 'prefix')
                          for ns_uri in (('id',) if ns_prefix else (None, '', 'id'))
                          if not is_content or not ns_uri)
    def test_attribute(self, value, required, is_content, ns_prefix, ns_uri):
        attr = ItemElementAttribute(value=value, required=required, is_content=is_content,
                                    ns_prefix=ns_prefix, ns_uri=ns_uri)
        six.assertRegex(
            self,
            repr(attr),
            r'^ItemElementAttribute\(value={}, serializer=[^,]+, required={}, is_content={}, ns_prefix={}, ns_uri={}\)$'
            .format(re.escape(repr(value)), re.escape(repr(required)), re.escape(repr(is_content)),
                    re.escape(repr(ns_prefix or '')), re.escape(repr(ns_uri or ''))))

    @parameterized.expand((attr_name, attr, elem_kwargs)
                          for attr_name, attr in chain([("attr0", ItemElementAttribute())], ns_attrs.items())
                          for elem_kwargs in chain([{}], ns_elem_names.values()))
    def test_element_with_single_attr(self, attr_name, attr, elem_kwargs):
        elem_cls_name = "Element0"
        elem_cls = type(elem_cls_name, (ItemElement,), {attr_name: attr})
        elem = elem_cls(**elem_kwargs)
        full_elem_kwargs = {"ns_prefix": "", "ns_uri": ""}
        full_elem_kwargs.update(elem_kwargs)
        expected_kwargs_reprs = [", {}={!r}".format(k, v) for k, v in full_elem_kwargs.items()]
        elem_reprs = ["{}({}={!r}{})".format(elem_cls_name, attr_name, attr,
                                        "".join(expected_kwargs_repr))
                      for expected_kwargs_repr in permutations(expected_kwargs_reprs)]
        assert any(repr(elem) == elem_repr for elem_repr in elem_reprs),\
            "{!r}\nis not equal to one of:\n{}".format(elem, "\n".join(elem_reprs))


    @parameterized.expand(product(combinations(
        chain([("attr0", ItemElementAttribute())], ns_attrs.items()),
        3),
        chain([{}], ns_elem_names.values())))
    def test_element_with_multiple_attrs(self, attrs, elem_kwargs):
        attrs = dict(attrs)
        elem_cls_name = "Element0"
        elem_cls = type(elem_cls_name, (ItemElement,), deepcopy(attrs))
        elem = elem_cls(**elem_kwargs)
        full_elem_kwargs = {"ns_prefix": "", "ns_uri": ""}
        full_elem_kwargs.update(elem_kwargs)
        for attr_name, attr in attrs.items():
            if not attr.ns_prefix and "__" in attr_name:
                attr.ns_prefix = attr_name.split("__")[0]
        expected_kwargs_reprs = [", {}={!r}".format(k, v) for k, v in full_elem_kwargs.items()]
        attrs_reprs = ["{}={!r}".format(name, attr) for name, attr in attrs.items()]
        elem_reprs = ["{}({}{})".format(elem_cls_name,
                                        ", ".join(attrs_repr),
                                        "".join(expected_kwargs_repr))
                      for attrs_repr in permutations(attrs_reprs)
                      for expected_kwargs_repr in permutations(expected_kwargs_reprs)]
        assert any(repr(elem) == elem_repr for elem_repr in elem_reprs),\
            "{!r}\nis not equal to one of:\n{}".format(elem, "\n".join(elem_reprs))

    @parameterized.expand((attr_name, attr, elem_name, elem_kwargs)
                          for attr_name, attr in chain([("attr0", ItemElementAttribute())], ns_attrs.items())
                          for elem_name, elem_kwargs in chain([('elem0', {})], ns_elem_names.items()))
    def test_item_with_single_elem(self, attr_name, attr, elem_name, elem_kwargs):
        elem_cls_name = "Element0"
        item_cls_name = "Item0"
        elem_cls = type(elem_cls_name, (ItemElement,), {attr_name: attr})
        elem = elem_cls(**elem_kwargs)
        item_cls = type(item_cls_name, (RssItem,), {elem_name: elem})
        item = item_cls()
        repr(item)
        if sys.version_info >= (3, 7): # insertion ordered dict
            default_elems_repr = ("{}={!r}".format(name, value)
                                  for name, value in chain(RssItem().elements.items(),
                                                           [(elem_name, elem)]))
            assert repr(item) == "{}({})".format(item_cls_name,
                                                 ", ".join(default_elems_repr))

    @parameterized.expand(product(
        combinations(chain([("attr0", ItemElementAttribute())], ns_attrs.items()), 3),
        combinations(chain([("elem0", {})], ns_elem_names.items()), 3)))
    def test_item_with_multiple_elems(self, attrs, elems_descr):
        elems_names, elems_kwargs = zip(*elems_descr)
        item_cls_name = "Item0"
        elem_clses = [type("Element{}".format(n), (ItemElement,), dict(attrs))
                      for n in range(len(elems_descr))]
        elem_instances = [elem_cls(**elems_kwargs[n])
                          for n, elem_cls in enumerate(elem_clses)]
        item_cls = type(item_cls_name, (RssItem,), dict(zip(elems_names, elem_instances)))
        item = item_cls()
        repr(item)
        if sys.version_info >= (3, 7): # insertion ordered dict
            elems_reprs = ("{}={}".format(elem_name, elem)
                           for elem_name, elem in chain(RssItem().elements.items(),
                                                        zip(elems_names, elem_instances)))
            item_repr = "{}({})".format(item_cls_name, ", ".join(elems_reprs))
            assert repr(item) == item_repr


class TestSimpleElements(RssTestCase):
    def __init__(self, *args, **kwargs):
        super(TestSimpleElements, self).__init__(*args, **kwargs)

        self.empty_text = ""
        self.non_empty_title = "Non-empty title"
        self.non_empty_description = "Non-empty description"
        self.categories = ["first category name", "second category name",
                           "third category name", "fourth category name"]
        self.unescaped_title = "<b>Non-empty<br/> title</b>"
        self.unescaped_description = "<b>Non-empty description</b><img src='url'/>"

        self.item_with_empty_title_only = RssItem()
        self.item_with_empty_title_only.title = self.empty_text

        self.item_with_empty_description_only = RssItem()
        self.item_with_empty_description_only.description = self.empty_text

        self.item_with_title_only = RssItem()
        self.item_with_title_only.title = self.non_empty_title

        self.item_with_description_only = RssItem()
        self.item_with_description_only.description = self.non_empty_description

        self.item_with_single_category = RssItem()
        self.item_with_single_category.title = self.non_empty_title
        self.item_with_single_category.category = self.categories[0]

        self.item_with_2_categories = RssItem()
        self.item_with_2_categories.title = self.non_empty_title
        self.item_with_2_categories.category = self.categories[:2]

        self.item_with_3_categories = RssItem()
        self.item_with_3_categories.title = self.non_empty_title
        self.item_with_3_categories.category = self.categories[:3]

        self.item_with_4_categories = RssItem()
        self.item_with_4_categories.title = self.non_empty_title
        self.item_with_4_categories.category = self.categories[:4]

        self.item_with_unescaped_text = RssItem()
        self.item_with_unescaped_text.title = self.unescaped_title
        self.item_with_unescaped_text.description = self.unescaped_description

        self.guids = [{'guid': 'identifier 1', 'isPermaLink': False},
                      {'guid': 'identifier 2', 'isPermaLink': True},]

        self.items_with_guid = {0: [], 1: []}

        item_with_guid = RssItem()
        item_with_guid.title = self.non_empty_title
        item_with_guid.guid = self.guids[0]['guid']
        self.items_with_guid[0].append(item_with_guid)

        item_with_guid = RssItem()
        item_with_guid.title = self.non_empty_title
        item_with_guid.guid = self.guids[0]
        self.items_with_guid[0].append(item_with_guid)

        item_with_guid = RssItem()
        item_with_guid.title = self.non_empty_title
        item_with_guid.guid.guid = self.guids[0]['guid']
        self.items_with_guid[0].append(item_with_guid)

        item_with_guid = RssItem()
        item_with_guid.title = self.non_empty_title
        item_with_guid.guid = self.guids[1]
        self.items_with_guid[1].append(item_with_guid)

        item_with_guid = RssItem()
        item_with_guid.title = self.non_empty_title
        item_with_guid.guid = self.guids[1]['guid']
        item_with_guid.guid.isPermaLink = self.guids[1]['isPermaLink']
        self.items_with_guid[1].append(item_with_guid)

        item_with_guid = RssItem()
        item_with_guid.title = self.non_empty_title
        item_with_guid.guid = GuidElement(**self.guids[1])
        self.items_with_guid[1].append(item_with_guid)

    @parameterized.expand((elem, str(elem_name))
                          for elem_name, elem in RssItem().elements.items())
    def test_elements_uniqueness(self, elem, elem_name):
        elem1 = elem.__class__() if not isinstance(elem, MultipleElements) else elem.__class__(ItemElement)
        elem2 = elem.__class__() if not isinstance(elem, MultipleElements) else elem.__class__(ItemElement)
        self.assertIsNot(elem1, elem2,
                         msg="Instances of element class '{}' are identical".format(elem.__class__.__name__))

        item1 = RssItem()
        item2 = RssItem()
        self.assertIsNot(getattr(item1, elem_name), getattr(item2, elem_name),
                         msg="Appropriate elements [class '{}'] of RSS item instances are identical"
                             .format(elem.__class__.__name__))

    @parameterized.expand((elem, str(elem_name), attr, attr_name)
                          for elem_name, elem in RssItem().elements.items()
                          for attr_name, attr in elem.attrs.items())
    def test_attributes_uniqueness(self, elem, elem_name, attr, attr_name):
        item1 = RssItem()
        item2 = RssItem()
        attr1 = attr.__class__()
        attr2 = attr.__class__()
        self.assertIsNot(attr1, attr2,
                         msg="Instances of attribute [class '{}'] are identical"
                             .format(attr.__class__.__name__))

        self.assertIsNot(getattr(getattr(item1, elem_name), attr_name.priv_name),
                         getattr(getattr(item2, elem_name), attr_name.priv_name),
                         msg="Appropriate attributes [class '{}'] of appropriate elements [class '{}'] "
                             "of RSS item instances are identical"
                             .format(attr.__class__.__name__, elem.__class__.__name__))

    @parameterized.expand((elem, str(elem_name), value)
                          for elem_name, elem in RssItem().elements.items()
                          for value in values)
    def test_item_properties_v1(self, elem, elem_name, value):
        item = RssItem()
        if elem.required_attrs:
            with six.assertRaisesRegex(self, ValueError, 'Could not assign value'):
                setattr(item, elem_name, value)
        else:
            setattr(item, elem_name, value)
            self.assertEqual(getattr(item, elem_name), value)

    def test_item_properties_v2(self):
        self.assertEqual(self.item_with_empty_title_only.title, self.empty_text)

        self.assertEqual(self.item_with_empty_description_only.description, self.empty_text)

        self.assertEqual(self.item_with_title_only.title, self.non_empty_title)

        self.assertEqual(self.item_with_description_only.description, self.non_empty_description)

        self.assertEqual(self.item_with_single_category.title, self.non_empty_title)
        self.assertEqual(self.item_with_single_category.category, self.categories[0])

        self.assertEqual(self.item_with_3_categories.title, self.non_empty_title)
        self.assertEqual(self.item_with_2_categories.category, self.categories[:2])

        self.assertEqual(self.item_with_4_categories.title, self.non_empty_title)
        self.assertEqual(self.item_with_3_categories.category, self.categories[:3])

        self.assertEqual(self.item_with_4_categories.title, self.non_empty_title)
        self.assertEqual(self.item_with_4_categories.category, self.categories[:4])

        self.assertEqual(self.item_with_unescaped_text.title, self.unescaped_title)

        self.assertEqual(self.item_with_unescaped_text.description, self.unescaped_description)

        for idx, items in self.items_with_guid.items():
            for item in items:
                self.assertEqual(item.guid, self.guids[idx]['guid'])
                self.assertEqual(item.guid.guid, self.guids[idx]['guid'])
                self.assertEqual(item.guid.isPermaLink, self.guids[idx]['isPermaLink'])

    @parameterized.expand((elem,) for elem in RssItem().elements.values())
    def test_element_init_without_args(self, elem):
        elem_cls = elem.__class__
        if elem_cls is MultipleElements:
            elem_cls(ItemElement)
        else:
            elem_cls()

    @parameterized.expand((elem, str(attr), value)
                          for elem in RssItem().elements.values()
                          for attr in elem.attrs
                          for value in values
                          if not isinstance(elem, MultipleElements))
    def test_element_init_with_single_kwarg(self, elem, attr_name, value):
        elem_cls = elem.__class__
        elem_cls(**{attr_name: value})

    @parameterized.expand((elem, str(bad_attr), value)
                          for elem in RssItem().elements.values()
                          for bad_attr in chain(('impossible_attr',),
                                                set(attr for elem in RssItem().elements.values() for attr in elem.attrs)
                                                - set(elem.attrs))
                          for value in values
                          if not isinstance(elem, MultipleElements))
    def test_element_init_with_bad_kwarg(self, elem, bad_attr_name, value):
        elem_cls = elem.__class__
        with six.assertRaisesRegex(self, ValueError, 'supports only the next named arguments',
                                     msg="Invalid attribute '{}' was passed to '{}' initializer"
                                         .format(bad_attr_name, elem_cls.__name__)):
            elem_cls(**{bad_attr_name: value})

    @parameterized.expand((elem, value)
                          for elem in RssItem().elements.values()
                          for value in values
                          if not isinstance(elem, MultipleElements))
    def test_element_init_content_arg(self, elem, value):
        elem_cls = elem.__class__
        if elem.content_arg:
            el = elem_cls(value)
            self.assertEqual(el, getattr(el, str(el.content_arg)))
            self.assertEqual(el, value)
        else:
            with six.assertRaisesRegex(self, ValueError, 'does not support unnamed arguments',
                                         msg="Invalid attribute was passed to '{}' initializer "
                                             "(element must not have content)".format(elem_cls.__name__)):
                elem_cls(value)

    @parameterized.expand((elem, value1, value2)
                          for elem in RssItem().elements.values()
                          for value1, value2 in zip(values, values)
                          if not isinstance(elem, MultipleElements))
    def test_element_init_with_multiple_args(self, elem, value1, value2):
        elem_cls = elem.__class__
        if elem.content_arg:
            with six.assertRaisesRegex(self, ValueError, 'supports only single unnamed argument',
                                         msg="Invalid attribute was passed to '{}' initializer "
                                             "(element must not have content)".format(elem_cls.__name__)):
                elem_cls(value1, value2)
        else:
            with six.assertRaisesRegex(self, ValueError, 'does not support unnamed arguments',
                                         msg="Invalid attribute was passed to '{}' initializer "
                                             "(element must not have content)".format(elem_cls.__name__)):
                elem_cls(value1, value2)


    @parameterized.expand((str(elem_name), str(attr_name), value)
                          for elem_name, elem_descr in RssItem().elements.items()
                          for attr_name in elem_descr.attrs
                          for value in values)
    def test_element_setattr(self, elem_name, attr_name, value):
        item = RssItem()
        elem = getattr(item, elem_name)
        setattr(elem, attr_name, value)
        self.assertEqual(getattr(elem, attr_name), value)

    def test_multi_content_element(self):
        with six.assertRaisesRegex(self, ValueError, r"More than one attributes.*as content"):
            class Element0(ItemElement):
                attr1 = ItemElementAttribute(is_content=True)
                attr2 = ItemElementAttribute(is_content=False)
                attr3 = ItemElementAttribute(is_content=True)


class TestMultipleElements(RssTestCase):
    def __init__(self, *args, **kwargs):
        super(TestMultipleElements, self).__init__(*args, **kwargs)
        self.category_names = ['1st category name', '2nd category name', '3rd category name', '4th category']
        non_empty_title = 'Item title'
        self.item_with_single_category = RssItem()
        self.item_with_single_category.title = non_empty_title
        self.item_with_single_category.category = self.category_names[0]

        self.item_with_2_categories = RssItem()
        self.item_with_2_categories.title = non_empty_title
        self.item_with_2_categories.category = self.category_names[:2]

        self.item_with_3_categories = RssItem()
        self.item_with_3_categories.title = non_empty_title
        self.item_with_3_categories.category = self.category_names[:3]

        self.item_with_4_categories = RssItem()
        self.item_with_4_categories.title = non_empty_title
        self.item_with_4_categories.category = self.category_names[:4]

    def test_methods(self):
        me = MultipleElements(ItemElement)
        self.assertEqual(len(me), 0)
        self.assertFalse(me.assigned)
        with self.assertRaises(NotImplementedError):
            me.serialize()
        with self.assertRaises(NotImplementedError):
            MultipleElements(CategoryElement).serialize()

        new_item1 = ItemElement()
        new_item2 = ItemElement()

        me.append(new_item1)
        self.assertTrue(me.assigned)
        self.assertEqual(len(me), 1)
        self.assertIs(me[0], new_item1)

        self.assertIs(me.pop(), new_item1)
        self.assertFalse(me.assigned)
        self.assertEqual(len(me), 0)

        me.extend([new_item1, new_item2])
        self.assertTrue(me.assigned)
        self.assertEqual(len(me), 2)
        self.assertIs(me[0], new_item1)
        self.assertIs(me[1], new_item2)

        self.assertIs(me.pop(), new_item2)
        self.assertTrue(me.assigned)
        self.assertEqual(len(me), 1)

        self.assertIs(me.pop(), new_item1)
        self.assertFalse(me.assigned)
        self.assertEqual(len(me), 0)

        me.add(new_item1)
        self.assertTrue(me.assigned)
        self.assertEqual(len(me), 1)

        self.assertIs(me.pop(), new_item1)
        self.assertFalse(me.assigned)
        self.assertEqual(len(me), 0)

        me.add([new_item1, new_item2])
        self.assertTrue(me.assigned)
        self.assertEqual(len(me), 2)

        self.assertIs(me.pop(), new_item2)
        self.assertTrue(me.assigned)
        self.assertEqual(len(me), 1)

        self.assertIs(me.pop(), new_item1)
        self.assertFalse(me.assigned)
        self.assertEqual(len(me), 0)

        me.add([new_item1, new_item2])
        self.assertIs(me.pop(0), new_item1)
        self.assertTrue(me.assigned)
        self.assertEqual(len(me), 1)

        self.assertIs(me.pop(0), new_item2)
        self.assertFalse(me.assigned)
        self.assertEqual(len(me), 0)

        me.add([new_item1, new_item2])
        del me[0]
        self.assertTrue(me.assigned)
        self.assertEqual(len(me), 1)
        self.assertIs(me[0], new_item2)

        del me[0]
        self.assertFalse(me.assigned)
        self.assertEqual(len(me), 0)

        me.add(new_item1)
        with self.assertRaises(TypeError):
            me[0] = type('some_class')()
        self.assertTrue(me.assigned)
        self.assertEqual(len(me), 1)
        me[0] = new_item2
        self.assertTrue(me.assigned)
        self.assertEqual(len(me), 1)
        self.assertIs(me[0], new_item2)
        with self.assertRaises(IndexError):
            me[1] = new_item1

    def test_initializer(self):
        with self.assertRaises(TypeError):
            MultipleElements(type('some_non_element_class'))

        me = MultipleElements(CategoryElement)
        me.add({'category': self.category_names[0]})
        self.assertEqual(me.category, self.category_names[0])

        me = MultipleElements(CategoryElement)
        me.add([{'category': cat_name} for cat_name in self.category_names])
        self.assertSequenceEqual([el.category for el in me], self.category_names)

    def test_instances(self):
        self.assertIsInstance(self.item_with_single_category.category, MultipleElements)
        for category in self.item_with_single_category.category:
            self.assertIsInstance(category, CategoryElement)
        self.assertIsInstance(self.item_with_2_categories.category, MultipleElements)
        for category in self.item_with_2_categories.category:
            self.assertIsInstance(category, CategoryElement)
        self.assertIsInstance(self.item_with_3_categories.category, MultipleElements)
        for category in self.item_with_3_categories.category:
            self.assertIsInstance(category, CategoryElement)
        self.assertIsInstance(self.item_with_4_categories.category, MultipleElements)
        for category in self.item_with_4_categories.category:
            self.assertIsInstance(category, CategoryElement)

    def test_inner_cls_attr(self):
        item = RssItem()
        item.category = CategoryElement(category=self.category_names[0])
        self.assertEqual(item.category, self.category_names[0])
        for cnt in range(2, 2+len(self.category_names)):
            item.category = [CategoryElement(category=cat_name)
                             for cat_name in self.category_names[:cnt]]
            self.assertEqual(item.category, self.category_names[:cnt])

            item.category = CategoryElement(category=self.category_names[0])
            self.assertEqual(item.category, self.category_names[0])

            item.category = self.category_names[:cnt]
            self.assertEqual(item.category, self.category_names[:cnt])

            item.category = self.category_names[0]
            self.assertEqual(item.category, self.category_names[0])

    def test_irregular_access(self):
        me = MultipleElements(CategoryElement)
        with six.assertRaisesRegex(self, AttributeError, 'have not been assigned'):
            me.category
        me.add(['first', 'second'])
        with six.assertRaisesRegex(self, AttributeError, 'Cannot get attribute: more than one elements'):
            me.category
        with six.assertRaisesRegex(self, AttributeError, 'Cannot set attribute: 2 elements have been assigned'):
            me.category = 'another'

        me.clear()
        me.add('single')
        me.category = 'another'
        self.assertEqual(me.category, 'another')

        item = RssItem()
        with six.assertRaisesRegex(self, AttributeError, 'have not been assigned'):
            item.category.category
        item.category = ['first', 'second']
        with six.assertRaisesRegex(self, AttributeError, 'Cannot get attribute: more than one elements'):
            item.category.category
        with six.assertRaisesRegex(self, AttributeError, 'Cannot set attribute: 2 elements have been assigned'):
            item.category.category = 'other'

        item.category.clear()
        item.category = 'single'
        item.category.category = 'another'
        self.assertEqual(item.category.category, 'another')


class TestNamespacedElements(RssTestCase):
    @parameterized.expand((elem_name, ns_kwargs, attr_name, attr)
                          for elem_name, ns_kwargs in ns_elem_names.items()
                          for attr_name, attr in ns_attrs.items())
    def test_access_by_name(self, elem_name, ns_kwargs, attr_name, attr):
        elem_cls = type("Element0", (ItemElement,), {attr_name: attr})
        elem = elem_cls(**ns_kwargs)
        item_cls = type("Item0", (RssItem,), {elem_name: elem})
        item = item_cls()
        actual_elem = getattr(item, elem_name)
        self.assertIsInstance(actual_elem, elem_cls)
        self.assertIsNot(actual_elem, elem)
        actual_attr_value = getattr(actual_elem, attr_name)
        self.assertIsNone(actual_attr_value, msg="Unexpected attribute value")

        value = "string"
        setattr(actual_elem, attr_name, value)
        actual_attr_value = getattr(actual_elem, attr_name)
        self.assertIs(actual_attr_value, value, msg="Unexpected attribute value")


    @parameterized.expand((elem_name, attr_name, attr)
                          for elem_name in ns_elem_names
                          for attr_name, attr in ns_attrs.items())
    def test_attr_namespace(self, elem_name, attr_name, attr):
        elem_cls = type("Element0", (ItemElement,), {attr_name: attr})
        elem = elem_cls()
        actual_attr = next(iter(elem.attrs.values()))
        self.assertIs(actual_attr, getattr(elem, "__" + attr_name))
        self.assertEqual(actual_attr.ns_prefix,
                         attr.ns_prefix or attr_name.split("__")[0])
        self.assertEqual(actual_attr.ns_uri, attr.ns_uri)


    @parameterized.expand(zip(ns_elem_names.values()))
    def test_content_attr_namespace(self, ns_kwargs):
        with six.assertRaisesRegex(self, ValueError, "Content cannot have namespace"):
            ItemElementAttribute(is_content=True, **ns_kwargs)


    @parameterized.expand((elem_name, ns_kwargs, attr_name, attr)
                          for elem_name, ns_kwargs in ns_elem_names.items()
                          for attr_name, attr in ns_attrs.items())
    def test_elem_namespace(self, elem_name, ns_kwargs, attr_name, attr):
        elem_cls = type("Element0", (ItemElement,), {attr_name: attr})
        elem = elem_cls(**ns_kwargs)
        item_cls = type("Item0", (RssItem,), {elem_name: elem})
        item = item_cls()
        actual_elem = next(v for n, v in item.elements.items() if str(n) == elem_name)
        self.assertIs(actual_elem, getattr(item, elem_name))
        self.assertEqual(actual_elem.ns_prefix,
                         ns_kwargs.get("ns_prefix", elem_name.split("__")[0]))
        self.assertEqual(actual_elem.ns_uri, ns_kwargs["ns_uri"])


    @parameterized.expand(combinations(ns_attrs.values(), 1))
    def test_attr_get_namespaces(self, attr):
        actual_namespaces = attr.get_namespaces()
        self.assertIsInstance(actual_namespaces, set)
        self.assertEqual(actual_namespaces, {(attr.ns_prefix, attr.ns_uri)})
        
        actual_namespaces = attr.get_namespaces(False)
        self.assertIsInstance(actual_namespaces, set)
        self.assertEqual(actual_namespaces, {(attr.ns_prefix, attr.ns_uri)})


    @parameterized.expand((elem_name, ns_kwargs, deepcopy(attrs))
                          for elem_name, ns_kwargs in ns_elem_names.items()
                          for attrs in chain(combinations(ns_attrs.items(), 1),
                                             combinations(ns_attrs.items(), 2),
                                             combinations(ns_attrs.items(), 3)))
    def test_elem_get_namespaces_with_value(self, elem_name, ns_kwargs, attrs):
        elem_cls = type("Element0", (ItemElement,), dict(attrs))
        elem_kwargs = ns_kwargs.copy()
        elem_kwargs.update({str(attr_name): "" for attr_name, _ in attrs})
        elem = elem_cls(**elem_kwargs)
        expected_namespaces = {(attr.ns_prefix, attr.ns_uri) for _, attr in attrs}
        expected_namespaces.add((ns_kwargs.get("ns_prefix", ""), ns_kwargs["ns_uri"]))
        actual_namespaces = elem.get_namespaces()
        self.assertIsInstance(actual_namespaces, set)
        self.assertEqual(actual_namespaces, expected_namespaces)


    @parameterized.expand((elem_name, ns_kwargs, deepcopy(attrs))
                          for elem_name, ns_kwargs in ns_elem_names.items()
                          for attrs in chain(combinations(ns_attrs.items(), 1),
                                             combinations(ns_attrs.items(), 2),
                                             combinations(ns_attrs.items(), 3)))
    def test_elem_get_namespaces_without_value(self, elem_name, ns_kwargs, attrs):
        elem_cls = type("Element0", (ItemElement,), dict(attrs))
        elem = elem_cls(**ns_kwargs)
        actual_namespaces = elem.get_namespaces()
        expected_namespaces = {(ns_kwargs.get("ns_prefix", ""), ns_kwargs["ns_uri"])}
        self.assertIsInstance(actual_namespaces, set)
        self.assertEqual(actual_namespaces, expected_namespaces)

        actual_namespaces = elem.get_namespaces(False)
        expected_namespaces.update({(attr.ns_prefix, attr.ns_uri) for _, attr in attrs})
        self.assertIsInstance(actual_namespaces, set)
        self.assertEqual(actual_namespaces, expected_namespaces)


    def test_item_get_namespaces(self):
        class Element0(ItemElement):
            attr01 = ItemElementAttribute(ns_prefix="prefix01", ns_uri="id01")

        class Element1(ItemElement):
            prefix11__attr11 = ItemElementAttribute(ns_uri="id11")
            prefix12__attr12 = ItemElementAttribute(ns_prefix="prefix12", ns_uri="id12")

        class Element2(ItemElement):
            attr21 = ItemElementAttribute(ns_prefix="prefix21", ns_uri="id01")
            pseudo_prefix22__attr22 = ItemElementAttribute(ns_prefix="prefix22", ns_uri="id22")
        
        class Item0(RssItem):
            elem0 = ItemElement()
            elem1 = Element0(ns_prefix="el_prefix1", ns_uri="el_id1")
            el_prefix2__elem2 = Element1(ns_uri="el_id2")
            el_prefix3__elem3 = Element2(ns_prefix="el_prefix3", ns_uri="el_id3")
            el_pseudo_prefix4__elem4 = Element0(ns_prefix="el_prefix4", ns_uri="el_id4")

        item1 = Item0()
        item1.elem1.attr01 = ""
        item1.el_prefix2__elem2.prefix11__attr11 = 0
        item1.el_prefix2__elem2.prefix12__attr12 = ""
        item1.el_prefix3__elem3.attr21 = ""
        item1.el_prefix3__elem3.pseudo_prefix22__attr22 = 0
        item1.el_pseudo_prefix4__elem4.attr01 = ""

        expected_namespaces = {("el_prefix1", "el_id1"), ("prefix01", "id01"),
                               ("el_prefix2", "el_id2"), ("prefix11", "id11"), ("prefix12", "id12"),
                               ("el_prefix3", "el_id3"), ("prefix21", "id01"), ("prefix22", "id22"),
                               ("el_prefix4", "el_id4"), ("prefix01", "id01")}
        actual_namespaces = item1.get_namespaces()
        self.assertIsInstance(actual_namespaces, set)
        self.assertEqual(actual_namespaces, expected_namespaces)

        item0 = Item0()
        actual_namespaces = item0.get_namespaces()
        self.assertIsInstance(actual_namespaces, set)
        self.assertEqual(actual_namespaces, set())

        actual_namespaces = item0.get_namespaces(False)
        self.assertIsInstance(actual_namespaces, set)
        self.assertEqual(actual_namespaces, expected_namespaces)


    def test_item_get_namespaces_with_multiplied_elems(self):
        class Element0(ItemElement):
            attr00 = ItemElementAttribute(is_content=True)

        class Element1(ItemElement):
            attr10 = ItemElementAttribute(ns_prefix="prefix10", ns_uri="id10")
        
        class Item0(RssItem):
            elem0 = ItemElement()
            elem1 = Element1(ns_prefix="el_prefix1", ns_uri="el_id1")
            el_prefix2__elem2 = MultipleElements(Element0, ns_uri="el_id2")
            el_prefix3__elem3 = MultipleElements(Element1, ns_prefix="el_prefix3", ns_uri="el_id3")
            el_pseudo_prefix4__elem4 = MultipleElements(Element0, ns_prefix="el_prefix4", ns_uri="el_id4")
            elem5 = MultipleElements(Element0, ns_prefix="el_prefix5", ns_uri="el_id5")

        
        item = Item0()
        item.elem1.attr10 = ""
        expected_namespaces = {("el_prefix1", "el_id1"), ("prefix10", "id10")}
        actual_namespaces = item.get_namespaces()
        self.assertIsInstance(actual_namespaces, set)
        self.assertEqual(actual_namespaces, expected_namespaces)


        item = Item0()
        item.el_prefix2__elem2.append("value")
        item.el_prefix3__elem3.extend([{"attr10": "value1"}, {"attr10": "value2"}])
        item.el_pseudo_prefix4__elem4.extend(["value1",
                                              Element0(attr00="value21", ns_uri="el_id4"),
                                              Element0(attr00="value22", ns_prefix="el_prefix4", ns_uri="el_id4"),
                                              {"attr00": "value31", "ns_uri": "el_id4"},
                                              {"attr00": "value32", "ns_prefix": "el_prefix4"},
                                              {"attr00": "value33", "ns_prefix": "el_prefix4", "ns_uri": "el_id4"}])
        item.elem5.extend(["value1", Element0(attr00="value2"), {"attr00": "value3"}])
        expected_namespaces = {("prefix10", "id10"),
                               ("el_prefix2", "el_id2"), ("el_prefix3", "el_id3"),
                               ("el_prefix4", "el_id4"), ("el_prefix5", "el_id5")}
        actual_namespaces = item.get_namespaces()
        self.assertIsInstance(actual_namespaces, set)
        self.assertEqual(actual_namespaces, expected_namespaces)


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-

from .nscomponent import BaseNSComponent, NSComponentName
from .attribute import ItemElementAttribute
from .element import ItemElementMeta, ItemElement, MultipleElements
from .item import ItemMeta, FeedItem, ExtendableItem



# -*- coding: utf-8 -*-

import os
from datetime import datetime
from itertools import chain, combinations
try:
    from tempfile import TemporaryDirectory
except ImportError:
    from backports.tempfile import TemporaryDirectory
from parameterized import parameterized
import six
from lxml import etree

import scrapy
from scrapy import signals
try:
    from scrapy.item import BaseItem
except ImportError:
    from scrapy.item import Item as BaseItem
from tests.utils import RaisedItemPipelineManager
from scrapy.exceptions import NotConfigured, CloseSpider
from scrapy.utils.misc import load_object
from scrapy.utils.test import get_crawler
from scrapy.core.scraper import Scraper
from scrapy.crawler import Crawler

from scrapy_rss.items import RssItem, FeedItem, RssedItem
from scrapy_rss.meta import ItemElement, ItemElementAttribute
from scrapy_rss.exceptions import *
from scrapy_rss.exporters import RssItemExporter

import unittest
from tests.utils import RssTestCase, FrozenDict


if six.PY2:
    import sys
    reload(sys)
    sys.setdefaultencoding('utf-8')


class CrawlerContext(object):
    default_settings = FrozenDict({'ITEM_PIPELINES':
                                       {'scrapy_rss.pipelines.RssExportPipeline': 900,},
                                   'LOG_LEVEL': 'WARNING',
                                   'EXTENSIONS': {
                                       'scrapy.extensions.memusage.MemoryUsage': None,},
                                   })

    def __init__(self, feed_file=None, feed_title=None, feed_link=None, feed_description=None,
                 crawler_settings=None):
        settings = crawler_settings if crawler_settings else dict(self.default_settings)
        if feed_file:
            settings['FEED_FILE'] = feed_file
        if feed_title:
            settings['FEED_TITLE'] = feed_title
        if feed_link:
            settings['FEED_LINK'] = feed_link
        if feed_description:
            settings['FEED_DESCRIPTION'] = feed_description
        self.crawler = get_crawler(settings_dict=settings)
        self.spider = scrapy.Spider.from_crawler(self.crawler, 'example.com')
        self.spider.parse = lambda response: ()
        item_processor = settings.get('ITEM_PROCESSOR')
        if not item_processor:
            item_processor = RaisedItemPipelineManager
        elif isinstance(item_processor, six.string_types):
            item_processor = load_object(item_processor)

        self.ipm = item_processor.from_crawler(self.crawler)

    def __enter__(self):
        responses = self.crawler.signals.send_catch_log(signal=signals.spider_opened,
                                                        spider=self.spider)
        for _, failure in responses:
            if failure:
                failure.raiseException()

        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        responses = self.crawler.signals.send_catch_log(signal=signals.spider_closed,
                                                        spider=self.spider, reason=None)
        for _, failure in responses:
            if failure:
                failure.raiseException()


class FullRssItemExporter(RssItemExporter):
    def __init__(self, file, channel_title, channel_link, channel_description,
                 language='en-US', copyright='Data', managing_editor='m@dot.com (Manager Name)',
                 webmaster='web@dot.com (Webmaster Name)',
                 pubdate=datetime(2000, 2, 1, 0, 10, 30), last_build_date=datetime(2000, 2, 1, 5, 10, 30),
                 category='some category', generator='tester',
                 docs='http://example.com/rss_docs', ttl=60,
                 *args, **kwargs):
        super(FullRssItemExporter, self) \
            .__init__(file, channel_title, channel_link, channel_description,
                      language=language, copyright=copyright, managing_editor=managing_editor,
                      webmaster=webmaster, pubdate=pubdate, last_build_date=last_build_date,
                      category=category, generator=generator,
                      docs=docs, ttl=ttl, *args, **kwargs)


default_feed_settings = FrozenDict({'feed_file': 'feed.rss',
                                    'feed_title': 'Title',
                                    'feed_link': 'http://example.com/feed',
                                    'feed_description': 'Description'})


class FeedSettings(TemporaryDirectory):
    def __enter__(self):
        dirname = super(FeedSettings, self).__enter__()
        feed_settings = dict(default_feed_settings)
        feed_settings['feed_file'] = os.path.join(dirname, feed_settings['feed_file'])
        feed_settings = FrozenDict(feed_settings)
        return feed_settings


class PredefinedItems(object):
    def __init__(self, *args, **kwargs):
        class NSElement0(ItemElement):
            attr01 = ItemElementAttribute(ns_prefix="prefix01", ns_uri="id01")

        class NSElement1(ItemElement):
            prefix11__attr11 = ItemElementAttribute(ns_uri="id11")
            prefix12__attr12 = ItemElementAttribute(ns_prefix="prefix12", ns_uri="id12")

        class NSElement2(ItemElement):
            attr21 = ItemElementAttribute(is_content=True)
            pseudo_prefix22__attr22 = ItemElementAttribute(ns_prefix="prefix22", ns_uri="id22")

        class NSElement3(ItemElement):
            attr31 = ItemElementAttribute(is_content=True)
            attr32 = ItemElementAttribute(ns_prefix="prefixa", ns_uri="id32")

        class NSElement4(ItemElement):
            attr41 = ItemElementAttribute()
            prefix42__attr41 = ItemElementAttribute(ns_uri="id42")
        
        class NSItem0(RssItem):
            elem0 = ItemElement()
            elem1 = NSElement0(ns_prefix="el_prefix1", ns_uri="el_id1")
            el_prefix2__elem2 = NSElement1(ns_uri="el_id2")
            el_prefix3__elem3 = NSElement2(ns_prefix="el_prefix3", ns_uri="el_id3")
            el_pseudo_prefix4__elem4 = NSElement0(ns_prefix="el_prefix4", ns_uri="el_id4")

        class NSItem1(RssItem):
            elem1 = NSElement0(ns_prefix="el_prefix1", ns_uri="el_id1")
            el_prefix__elem2 = NSElement1(ns_uri="el_id2")
            elem3 = NSElement2(ns_prefix="el_prefix", ns_uri="el_id3")
            el_pseudo_prefix4__elem4 = NSElement0(ns_prefix="el_prefix4", ns_uri="el_id4")

        class NSItem2(RssItem):
            elem1 = NSElement3(ns_prefix="prefix", ns_uri="el_id1")
            prefix__elem2 = NSElement3(ns_uri="el_id2")
            elem3 = NSElement3(ns_prefix="prefix", ns_uri="el_id3")
            el_pseudo_prefix4__elem4 = NSElement3(ns_prefix="prefix", ns_uri="el_id4")

        class NSItem3(RssItem):
            elem1 = NSElement3(ns_uri="el_id1")
            elem2 = NSElement3(ns_uri="el_id2")
            elem3 = NSElement3(ns_prefix="prefix", ns_uri="el_id3")
            el_pseudo_prefix4__elem3 = NSElement3(ns_prefix="prefix2", ns_uri="el_id4")
            elem4 = NSElement4()
            elem5 = NSElement4()

        PredefinedItems.NSItem0 = NSItem0
        PredefinedItems.NSItem1 = NSItem1
        PredefinedItems.NSItem2 = NSItem2
        PredefinedItems.NSItem3 = NSItem3

        minimal_item = RssItem()
        minimal_item.title = 'Title of minimal item'

        minimal_item2 = RssItem()
        minimal_item2.description = 'Description of minimal item'

        simple_item = RssItem()
        simple_item.title = 'Title of simple item'
        simple_item.description = 'Description of simple item'

        item_with_single_category = RssItem()
        item_with_single_category.title = 'Title of item with single category'
        item_with_single_category.category = 'Category 1'

        item_with_multiple_categories = RssItem()
        item_with_multiple_categories.title = 'Title of item with multiple categories'
        item_with_multiple_categories.category = ['Category 1', 'Category 2']

        item_with_guid = RssItem()
        item_with_guid.title = 'Title of item with guid'
        item_with_guid.guid = 'Identifier'

        item_with_unicode = RssItem()
        item_with_unicode.title = 'Title of item with unicode and special characters'
        item_with_unicode.description = "[Testing «ταБЬℓσ»: 1<2 & 4+1>3, now 20% off!]"

        item_with_enclosure = RssItem()
        item_with_enclosure.title = 'Title of item with enclosure'
        item_with_enclosure.enclosure.url = 'http://example.com/content'
        item_with_enclosure.enclosure.length = 0
        item_with_enclosure.enclosure.type = 'text/plain'

        item_with_unique_ns = NSItem0()
        item_with_unique_ns.title = "Title of item with unique namespaces"
        item_with_unique_ns.elem1.attr01 = ""
        item_with_unique_ns.el_prefix2__elem2.prefix11__attr11 = 0
        item_with_unique_ns.el_prefix2__elem2.prefix12__attr12 = ""
        item_with_unique_ns.el_prefix3__elem3.attr21 = "value3_21"
        item_with_unique_ns.el_prefix3__elem3.pseudo_prefix22__attr22 = 42
        item_with_unique_ns.el_pseudo_prefix4__elem4.attr01 = ""

        item_with_non_unique_ns = NSItem1()
        item_with_non_unique_ns.title = "Title of item with unique namespaces"
        item_with_non_unique_ns.elem1.attr01 = "-"
        item_with_non_unique_ns.el_prefix__elem2.prefix11__attr11 = -1
        item_with_non_unique_ns.el_prefix__elem2.prefix12__attr12 = "-"
        item_with_non_unique_ns.elem3.attr21 = "yet another value3_21"
        item_with_non_unique_ns.elem3.pseudo_prefix22__attr22 = 4224
        item_with_non_unique_ns.el_pseudo_prefix4__elem4.attr01 = "-"

        item_with_non_unique_ns2 = NSItem1()
        item_with_non_unique_ns2.title = "Title of item with unique namespaces 2"
        item_with_non_unique_ns2.elem1.attr01 = "0"
        item_with_non_unique_ns2.el_prefix__elem2.prefix11__attr11 = -999
        item_with_non_unique_ns2.elem3.attr21 = "value"
        item_with_non_unique_ns2.elem3.pseudo_prefix22__attr22 = 42
        item_with_non_unique_ns2.el_pseudo_prefix4__elem4.attr01 = ""

        item_with_same_ns_prefixes = NSItem2()
        item_with_same_ns_prefixes.title = "Title of item with same namespace prefixes"
        item_with_same_ns_prefixes.elem1.attr31 = "Content value 11ё"
        item_with_same_ns_prefixes.prefix__elem2.attr32 = "Attribute value 22"
        item_with_same_ns_prefixes.elem3.attr31 = "Content value 11"
        item_with_same_ns_prefixes.elem3.attr32 = "Attribute value 32"
        item_with_same_ns_prefixes.el_pseudo_prefix4__elem4.attr32 = ""

        item_with_default_nses = NSItem3()
        item_with_default_nses.title = "Title of item with default namespaces"
        item_with_default_nses.elem1.attr31 = "Content value 11ё"
        item_with_default_nses.elem2.attr32 = "Attribute value 22"
        item_with_default_nses.elem3.attr31 = "Content value 11"
        item_with_default_nses.elem3.attr32 = "Attribute value 32"
        item_with_default_nses.el_pseudo_prefix4__elem3.attr32 = ""
        item_with_default_nses.elem4.attr41 = "A41 b"
        item_with_default_nses.elem4.prefix42__attr41 = "0"

        self.items = {
             'minimal_item': minimal_item,
             'minimal_item2': minimal_item2,
             'simple_item': simple_item,
             'item_with_single_category': item_with_single_category,
             'item_with_multiple_categories': item_with_multiple_categories,
             'item_with_guid': item_with_guid,
             'item_with_unicode': item_with_unicode,
             'item_with_enclosure': item_with_enclosure,
             'item_with_unique_ns': item_with_unique_ns,
             'item_with_non_unique_ns': item_with_non_unique_ns,
             'item_with_same_ns_prefixes': item_with_same_ns_prefixes,
             'item_with_default_nses': item_with_default_nses}

        self.ns_items_of_same_cls = [
            ('item_with_non_unique_ns5', NSItem1, item_with_non_unique_ns),
            ('item_with_non_unique_ns4', NSItem1, item_with_non_unique_ns2),
        ]
        self.ns_items = [
             ('item_with_unique_ns2',
              [("el_prefix1", "el_id1"), ("prefix01", "id01"), ("el_prefix2", "el_id2"), ("prefix11", "id11"), ("prefix12", "id12")],
              None,
              item_with_unique_ns),
             ('item_with_unique_ns2',
              (("el_prefix1", "el_id1"), ("prefix01", "id01"), ("el_prefix2", "el_id2"), ("prefix11", "id11"), ("prefix12", "id12")),
              tuple(),
              item_with_unique_ns),
             ('item_with_unique_ns2',
              {"el_prefix1": "el_id1","prefix01": "id01", "el_prefix2": "el_id2", "prefix11": "id11", "prefix12": "id12"},
              None,
              item_with_unique_ns),
             ('item_with_unique_ns3', None, NSItem0, item_with_unique_ns),
             ('item_with_unique_ns3',
              None, 
              'tests.test_exporter.NSItem0',
              item_with_unique_ns),
             ('item_with_non_unique_ns2',
              [("el_prefix1", "el_id1"), ("prefix01", "id01"), ("prefix11", "id11"), ("prefix12", "id12"), ("prefix22", "id22"), ("el_prefix4", "el_id4")],
              None,
              item_with_non_unique_ns),
             ('item_with_non_unique_ns3',
              {"el_prefix1": "el_id1", "prefix01": "id01", "prefix11": "id11", "prefix12": "id12", "prefix22": "id22"},
              None,
              item_with_non_unique_ns),
             ('item_with_non_unique_ns2', None, NSItem1, item_with_non_unique_ns),
             ('item_with_non_unique_ns2',
              None,
              'tests.test_exporter.NSItem1',
              item_with_non_unique_ns),
             ('item_with_same_ns_prefixes2',
              [("prefix", "el_id1"), ("prefixa", "id32"), ("unused_prefix", "id000")],
              None,
              item_with_same_ns_prefixes),
             ('item_with_same_ns_prefixes2',
              {"prefix": "el_id1", "prefixa": "id32", "unused_prefix": "id000"},
              None,
              item_with_same_ns_prefixes),
             ('item_with_same_ns_prefixes3', None, NSItem2, item_with_same_ns_prefixes),
             ('item_with_same_ns_prefixes3',
              None,
              'tests.test_exporter.NSItem2',
              item_with_same_ns_prefixes),
             ('item_with_default_nses3',
              {'prefixa': 'id32', 'prefix2': 'el_id4'},
              None,
              item_with_default_nses),
             ('item_with_default_nses2',
              None,
              'tests.test_exporter.NSItem3',
              item_with_default_nses),
             ('item_with_default_nses2',
              None,
              NSItem3,
              item_with_default_nses)
        ]


predefined_items = PredefinedItems()
NSItem0 = PredefinedItems.NSItem0
NSItem1 = PredefinedItems.NSItem1
NSItem2 = PredefinedItems.NSItem2
NSItem3 = PredefinedItems.NSItem3


class TestExporting(RssTestCase):
    @parameterized.expand(zip(chain.from_iterable(
        combinations(default_feed_settings.items(), r)
        for r in range(1, len(default_feed_settings)))))
    def test_partial_required_settings(self, partial_settings):
        with FeedSettings() as feed_settings:
            partial_settings = dict(partial_settings)
            if 'feed_file' in partial_settings:
                partial_settings['feed_file'] = feed_settings['feed_file']
            undefined_settings = [name.upper()
                                  for name in set(default_feed_settings) - set(partial_settings)]
            with six.assertRaisesRegex(self, NotConfigured,
                                       '({})'.format('|'.join(undefined_settings))
                                            if len(undefined_settings) > 1 else undefined_settings[0],
                                       msg='The feed file, title, link and description must be specified, but the absence of {} is allowed'
                                             .format(undefined_settings)):
                with CrawlerContext(**partial_settings):
                    pass

    def test_empty_feed(self):
        with self.assertRaises(CloseSpider):
            feed_settings = dict(default_feed_settings)
            feed_settings['feed_file'] = 'non/existent/filepath'
            with CrawlerContext(**feed_settings):
                pass

        with FeedSettings() as feed_settings:
            with CrawlerContext(**feed_settings):
                pass

            with open(feed_settings['feed_file']) as data, \
                 open(os.path.join(os.path.dirname(__file__), 'expected_rss', 'empty_feed.rss')) as expected:
                self.assertUnorderedXmlEquivalentOutputs(data.read(), expected.read())

    def test_custom_exporters(self):
        with FeedSettings() as feed_settings:
            crawler_settings = dict(CrawlerContext.default_settings)
            crawler_settings['FEED_EXPORTER'] = 'tests.test_exporter.FullRssItemExporter'

            with CrawlerContext(crawler_settings=crawler_settings, **feed_settings):
                pass
            with open(feed_settings['feed_file']) as data, \
                 open(os.path.join(os.path.dirname(__file__),
                                   'expected_rss', 'full_empty_feed.rss')) as expected:
                self.assertUnorderedXmlEquivalentOutputs(data.read(), expected.read(),
                                                         excepted_elements=None)

            crawler_settings['FEED_EXPORTER'] = FullRssItemExporter
            with CrawlerContext(crawler_settings=crawler_settings, **feed_settings):
                pass
            with open(feed_settings['feed_file']) as data, \
                 open(os.path.join(os.path.dirname(__file__),
                                   'expected_rss', 'full_empty_feed.rss')) as expected:
                self.assertUnorderedXmlEquivalentOutputs(data.read(), expected.read(),
                                                         excepted_elements=None)

            class InvalidRssItemExporter1(RssItemExporter):
                def __init__(self, file, channel_title, channel_link, channel_description,
                             managing_editor='Manager Name',
                             *args, **kwargs):
                    super(InvalidRssItemExporter1, self) \
                        .__init__(file, channel_title, channel_link, channel_description,
                                  managing_editor=managing_editor, *args, **kwargs)

            crawler_settings['FEED_EXPORTER'] = InvalidRssItemExporter1
            with six.assertRaisesRegex(self, ValueError, 'managing_editor'):
                with CrawlerContext(crawler_settings=crawler_settings, **feed_settings):
                    pass

            class InvalidRssItemExporter2(RssItemExporter):
                def __init__(self, file, channel_title, channel_link, channel_description,
                             webmaster='Webmaster Name',
                             *args, **kwargs):
                    super(InvalidRssItemExporter2, self) \
                        .__init__(file, channel_title, channel_link, channel_description,
                                  webmaster=webmaster, *args, **kwargs)

            crawler_settings['FEED_EXPORTER'] = InvalidRssItemExporter2
            with six.assertRaisesRegex(self, ValueError, 'webmaster'):
                with CrawlerContext(crawler_settings=crawler_settings, **feed_settings):
                    pass

            class MultipleCategoriesRssItemExporter(RssItemExporter):
                def __init__(self, file, channel_title, channel_link, channel_description,
                             category=('category 1', 'category 2'),
                             *args, **kwargs):
                    super(MultipleCategoriesRssItemExporter, self) \
                        .__init__(file, channel_title, channel_link, channel_description,
                                  category=category, *args, **kwargs)

            crawler_settings['FEED_EXPORTER'] = MultipleCategoriesRssItemExporter
            with CrawlerContext(crawler_settings=crawler_settings, **feed_settings):
                pass
            with open(feed_settings['feed_file']) as data, \
                 open(os.path.join(os.path.dirname(__file__),
                                   'expected_rss', 'empty_feed_with_categories.rss')) as expected:
                self.assertUnorderedXmlEquivalentOutputs(data.read(), expected.read())

            class NoGeneratorRssItemExporter(RssItemExporter):
                def __init__(self, file, channel_title, channel_link, channel_description,
                             generator='',
                             *args, **kwargs):
                    super(NoGeneratorRssItemExporter, self) \
                        .__init__(file, channel_title, channel_link, channel_description,
                                  generator=generator, *args, **kwargs)

            crawler_settings['FEED_EXPORTER'] = NoGeneratorRssItemExporter
            with CrawlerContext(crawler_settings=crawler_settings, **feed_settings):
                pass
            with open(feed_settings['feed_file']) as data, \
                 open(os.path.join(os.path.dirname(__file__), 'expected_rss', 'empty_feed_without_generator.rss')) as expected:
                self.assertUnorderedXmlEquivalentOutputs(data.read(), expected.read(), excepted_elements='lastBuildDate')

            class InvalidExporter(object):
                pass

            crawler_settings['FEED_EXPORTER'] = InvalidExporter
            with six.assertRaisesRegex(self, TypeError, 'FEED_EXPORTER'):
                with CrawlerContext(crawler_settings=crawler_settings, **feed_settings):
                    pass

    def test_item_validation(self):
        with FeedSettings() as feed_settings:
            invalid_item = RssItem()
            invalid_item.enclosure.url = 'http://example.com/content'

            with six.assertRaisesRegex(self, InvalidRssItemAttributesError,
                                       r'required attributes .*? not set'):
                with CrawlerContext(**feed_settings) as context:
                    context.ipm.process_item(invalid_item, context.spider)

            class NonStandardElement(ItemElement):
                first_attribute = ItemElementAttribute(required=True, is_content=True)
                second_attribute = ItemElementAttribute(required=True)

            class NonStandardItem(RssItem):
                element = NonStandardElement()

            invalid_item = NonStandardItem()
            with six.assertRaisesRegex(self, InvalidElementValueError, 'Could not assign'):
                invalid_item.element = 'valid value'
            invalid_item.element.first_attribute = 'valid value'

            with six.assertRaisesRegex(self, InvalidRssItemAttributesError,
                                       r'required attributes .*? not set'):
                with CrawlerContext(**feed_settings) as context:
                    context.ipm.process_item(invalid_item, context.spider)

            class InvalidSuperItem1(FeedItem):
                pass

            class InvalidSuperItem2(FeedItem):
                field = scrapy.Field()

            class InvalidSuperItem3(FeedItem):
                rss = scrapy.Field()

            for invalid_item_cls in (InvalidSuperItem1, InvalidSuperItem2, InvalidSuperItem3):
                with six.assertRaisesRegex(self, InvalidRssItemError, "Item must have 'rss'"):
                    with CrawlerContext(**feed_settings) as context:
                        context.ipm.process_item(invalid_item_cls(), context.spider)

    @parameterized.expand(zip([scrapy.Item, BaseItem, dict]))
    def test_bad_item_cls(self, item_cls):
        with FeedSettings() as feed_settings:
            crawler_settings = dict(CrawlerContext.default_settings)
            crawler_settings['FEED_ITEM_CLASS'] = item_cls

            with six.assertRaisesRegex(self, ValueError, 'must be RssItem'):
                with CrawlerContext(crawler_settings=crawler_settings, **feed_settings):
                    pass

    @parameterized.expand(predefined_items.items.items())
    def test_single_item_in_the_feed(self, item_name, item):
        with FeedSettings() as feed_settings:
            class SuperItem(FeedItem):
                some_field = scrapy.Field()

                def __init__(self):
                    super(SuperItem, self).__init__()
                    self.rss = RssItem()

            with CrawlerContext(**feed_settings) as context:
                context.ipm.process_item(item, context.spider)
            with open(feed_settings['feed_file']) as data, \
                 open(os.path.join(os.path.dirname(__file__),
                                   'expected_rss', '{}.rss'.format(item_name))) as expected:
                self.assertUnorderedXmlEquivalentOutputs(data=data.read(), expected=expected.read())

            super_item = SuperItem()
            super_item.rss = item
            with CrawlerContext(**feed_settings) as context:
                context.ipm.process_item(super_item, context.spider)
            with open(feed_settings['feed_file']) as data, \
                 open(os.path.join(os.path.dirname(__file__),
                                   'expected_rss', '{}.rss'.format(item_name))) as expected:
                self.assertUnorderedXmlEquivalentOutputs(data=data.read(), expected=expected.read())

    @parameterized.expand(predefined_items.ns_items)
    def test_single_ns_item_in_the_feed(self, item_name, namespaces, item_cls, item):
        with FeedSettings() as feed_settings:
            class SuperItem(FeedItem):
                some_field = scrapy.Field()

                def __init__(self):
                    super(SuperItem, self).__init__()
                    self.rss = RssItem()

            crawler_settings = dict(CrawlerContext.default_settings)
            if namespaces is not None:
                crawler_settings['FEED_NAMESPACES'] = namespaces
            if item_cls is not None:
                crawler_settings['FEED_ITEM_CLASS'] = item_cls

            with CrawlerContext(crawler_settings=crawler_settings, **feed_settings) as context:
                context.ipm.process_item(item, context.spider)
            with open(feed_settings['feed_file']) as data, \
                 open(os.path.join(os.path.dirname(__file__),
                                   'expected_rss', '{}.rss'.format(item_name))) as expected:
                self.assertUnorderedXmlEquivalentOutputs(data=data.read(), expected=expected.read())

            super_item = SuperItem()
            super_item.rss = item
            with CrawlerContext(crawler_settings=crawler_settings, **feed_settings) as context:
                context.ipm.process_item(super_item, context.spider)
            with open(feed_settings['feed_file']) as data, \
                 open(os.path.join(os.path.dirname(__file__),
                                   'expected_rss', '{}.rss'.format(item_name))) as expected:
                self.assertUnorderedXmlEquivalentOutputs(data=data.read(), expected=expected.read())

    def test_all_items_in_the_single_feed(self):
        with FeedSettings() as feed_settings:
            with open(os.path.join(os.path.dirname(__file__),
                                   'expected_rss', 'empty_feed.rss'), 'rb') as feed_f:
                feed_tree = etree.fromstring(feed_f.read())
                feed_channel = feed_tree.xpath('//channel')[0]
                with CrawlerContext(**feed_settings) as context:
                    for item_name, item in predefined_items.items.items():
                        context.ipm.process_item(item, context.spider)
                        with open(os.path.join(os.path.dirname(__file__),
                                               'expected_rss', '{}.rss'.format(item_name)), 'rb') as item_f:
                            item_tree = etree.fromstring(item_f.read())
                            feed_channel.extend(item_tree.xpath('//item'))
                with open(feed_settings['feed_file']) as data:
                    self.assertUnorderedXmlEquivalentOutputs(data.read(), feed_tree)

    def test_ns_items_in_the_single_feed(self):
        with FeedSettings() as feed_settings:
            base_filename, item_cls, _ = predefined_items.ns_items_of_same_cls[0]
            with open(os.path.join(os.path.dirname(__file__),
                                   'expected_rss', '{}.rss'.format(base_filename)), 'rb') as feed_f:
                feed_tree = etree.fromstring(feed_f.read())
                feed_channel = feed_tree.xpath('//channel')[0]
                for item in list(feed_channel.xpath('./item')):
                    feed_channel.remove(item)
                crawler_settings = dict(CrawlerContext.default_settings)
                crawler_settings['FEED_ITEM_CLS'] = item_cls
                with CrawlerContext(crawler_settings=crawler_settings, **feed_settings) as context:
                    for item_name, item_cls, item in predefined_items.ns_items_of_same_cls:
                        context.ipm.process_item(item, context.spider)
                        with open(os.path.join(os.path.dirname(__file__),
                                               'expected_rss', '{}.rss'.format(item_name)), 'rb') as item_f:
                            item_tree = etree.fromstring(item_f.read())
                            feed_channel.extend(item_tree.xpath('//item'))
                with open(feed_settings['feed_file']) as data:
                    self.assertUnorderedXmlEquivalentOutputs(data.read(), feed_tree)


class TestScraper:
    class MySpider(scrapy.Spider):
        name = 'spider'

    def test_spider_output_handling(self):
        spider = self.MySpider()
        scraper = Scraper(Crawler(self.MySpider))
        scraper.open_spider(spider)
        scraper._process_spidermw_output(RssItem(), None, None, None)
        scraper._process_spidermw_output(NSItem0(), None, None, None)
        scraper._process_spidermw_output(NSItem1(), None, None, None)
        scraper._process_spidermw_output(NSItem2(), None, None, None)
        scraper._process_spidermw_output(FeedItem(), None, None, None)
        scraper._process_spidermw_output(RssedItem(), None, None, None)
        scraper.close_spider(spider)


if __name__ == '__main__':
    unittest.main()




# -*- coding: utf-8 -*-

import six
from scrapy import signals
from scrapy.exceptions import NotConfigured, CloseSpider
from scrapy.utils.misc import load_object

from .items import RssItem
from .exporters import RssItemExporter


class RssExportPipeline(object):
    def __init__(self):
        self.files = {}
        self.exporters = {}

    @classmethod
    def from_crawler(cls, crawler):
        pipeline = cls()
        crawler.signals.connect(pipeline.spider_opened, signals.spider_opened)
        crawler.signals.connect(pipeline.spider_closed, signals.spider_closed)
        return pipeline

    def spider_opened(self, spider):
        try:
            file = open(spider.settings.get('FEED_FILE'), 'wb')
        except TypeError:
            raise NotConfigured('FEED_FILE parameter does not string or does not exist')
        except (IOError, OSError) as e:
            raise CloseSpider('Cannot open file {}: {}'.format(spider.settings.get('FEED_FILE', None), e))
        self.files[spider] = file
        feed_title = spider.settings.get('FEED_TITLE')
        if not feed_title:
            raise NotConfigured('FEED_TITLE parameter does not exist')
        feed_link = spider.settings.get('FEED_LINK')
        if not feed_link:
            raise NotConfigured('FEED_LINK parameter does not exist')
        feed_description = spider.settings.get('FEED_DESCRIPTION')
        if feed_description is None:
            raise NotConfigured('FEED_DESCRIPTION parameter does not exist')

        item_cls = spider.settings.get('FEED_ITEM_CLASS', spider.settings.get('FEED_ITEM_CLS', RssItem))
        if isinstance(item_cls, six.string_types):
            item_cls = load_object(item_cls)

        namespaces = spider.settings.get('FEED_NAMESPACES', {})

        feed_exporter = spider.settings.get('FEED_EXPORTER', RssItemExporter)
        if isinstance(feed_exporter, six.string_types):
            feed_exporter = load_object(feed_exporter)
        if not issubclass(feed_exporter, RssItemExporter):
            raise TypeError("FEED_EXPORTER must be RssItemExporter or its subclass, not '{}'".format(feed_exporter))
        self.exporters[spider] = feed_exporter(file, feed_title, feed_link, feed_description,
                                               namespaces=namespaces, item_cls=item_cls)
        self.exporters[spider].start_exporting()

    def spider_closed(self, spider):
        self.exporters[spider].finish_exporting()
        file = self.files.pop(spider)
        file.close()

    def process_item(self, item, spider):
        self.exporters[spider].export_item(item)
        return item



# -*- coding: utf-8 -*-

from . import meta
from .utils import format_rfc822


class TitleElement(meta.ItemElement):
    title = meta.ItemElementAttribute(required=True, is_content=True)


class LinkElement(meta.ItemElement):
    link = meta.ItemElementAttribute(required=True, is_content=True)


class DescriptionElement(meta.ItemElement):
    description = meta.ItemElementAttribute(required=True, is_content=True)


class AuthorElement(meta.ItemElement):
    author = meta.ItemElementAttribute(required=True, is_content=True)


class CategoryElement(meta.ItemElement):
    category = meta.ItemElementAttribute(required=True, is_content=True)


class CommentsElement(meta.ItemElement):
    comments = meta.ItemElementAttribute(required=True, is_content=True)


class EnclosureElement(meta.ItemElement):
    url = meta.ItemElementAttribute(required=True)
    length = meta.ItemElementAttribute(required=True)
    type = meta.ItemElementAttribute(required=True)


class GuidElement(meta.ItemElement):
    isPermaLink = meta.ItemElementAttribute(required=False, serializer=lambda v: str(v).lower(), value=False)
    guid = meta.ItemElementAttribute(required=True, is_content=True)


class PubDateElement(meta.ItemElement):
    datetime = meta.ItemElementAttribute(required=True, serializer=format_rfc822, is_content=True)


class SourceElement(meta.ItemElement):
    url = meta.ItemElementAttribute(required=True)



# -*- coding: utf-8 -*-

from .elements import *
from . import meta
from .meta.item import FeedItem, ExtendableItem


class RssItem(FeedItem):
    title = TitleElement()
    link = LinkElement()
    description = DescriptionElement()
    author = AuthorElement()
    category = meta.MultipleElements(CategoryElement)
    comments = CommentsElement()
    enclosure = EnclosureElement()
    guid = GuidElement()
    pubDate = PubDateElement()
    source = SourceElement()


class RssedItem(FeedItem):
    def __init__(self, *args, **kwargs):
        super(RssedItem, self).__init__(*args, **kwargs)
        self.rss = RssItem()




# -*- coding: utf-8 -*-

from copy import deepcopy
import six
from scrapy.item import ItemMeta as BaseItemMeta, Item as BaseItem, MutableMapping

try:
    from scrapy.item import _BaseItemMeta
except ImportError:
    _BaseItemMeta = type

from ..exceptions import *
from .nscomponent import NSComponentName
from .element import ItemElement, MultipleElements


class ItemMeta(BaseItemMeta):
    def __new__(mcs, cls_name, cls_bases, cls_attrs):
        cls_attrs['_elements'] = {}
        for cls_base in reversed(cls_bases):
            if isinstance(cls_base, ItemMeta):
                cls_attrs['_elements'].update(cls_base._elements)
                for attr_name, attr_value in cls_base.__dict__.items():
                    if isinstance(attr_value, ItemElement):
                        cls_attrs[attr_name] = attr_value
        elements = {NSComponentName(elem_name, elem_descr.ns_prefix, elem_descr.ns_uri):
                    elem_descr for elem_name, elem_descr in cls_attrs.items()
                    if isinstance(elem_descr, ItemElement)}
        for elem_name, elem in elements.items():
            if not elem.ns_prefix:
                elem.ns_prefix = elem_name.ns_prefix
        cls_attrs['_elements'].update(elements)
        cls_attrs['elements'] = property(lambda self: self._elements)
        cls_attrs.update({elem_name.priv_name: elem_descriptor
                          for elem_name, elem_descriptor in elements.items()})
        cls_attrs.update({elem_name.pub_name: property(mcs.build_elem_getter(elem_name),
                                                       mcs.build_elem_setter(elem_name, elem_descr))
                          for elem_name, elem_descr in elements.items()})
        return super(ItemMeta, mcs).__new__(mcs, cls_name, cls_bases, cls_attrs)

    @staticmethod
    def build_elem_getter(elem_name):
        return lambda self: getattr(self, elem_name.priv_name)

    @staticmethod
    def build_elem_setter(elem_name, elem_descriptor):
        def setter(self, new_value):
            if isinstance(elem_descriptor, MultipleElements):
                multi_elem = getattr(self, elem_name.priv_name)
                multi_elem.clear()
                multi_elem.add(new_value)
            elif isinstance(new_value, elem_descriptor.__class__):
                setattr(self, elem_name.priv_name, new_value)
            elif isinstance(new_value, dict):
                setattr(self, elem_name.priv_name, elem_descriptor.__class__(**new_value))
            elif not elem_descriptor.required_attrs and elem_descriptor.content_arg:
                elem = getattr(self, elem_name.priv_name)
                setattr(elem, elem_descriptor.content_arg.pub_name, new_value)
            else:
                raise InvalidElementValueError(elem_name, elem_descriptor.__class__, new_value)
        return setter


class FeedItem(six.with_metaclass(ItemMeta, BaseItem)):
    """
    Attributes
    ----------
    elements : { NSComponentName : ItemElement }
        All elements of the item
    """
    def __init__(self, *args, **kwargs):
        super(FeedItem, self).__init__(*args, **kwargs)
        new_elements = {}
        for elem_name, elem_descr in self._elements.items():
            new_element = deepcopy(getattr(self, elem_name.priv_name))
            setattr(self, elem_name.priv_name, new_element)
            new_elements[elem_name] = new_element
        self._elements = new_elements

    def __repr__(self):
        return "{}({})".format(
            self.__class__.__name__,
            ", ".join("{}={!r}".format(elem_name, elem)
                      for elem_name, elem in self.elements.items()))

    def __setattr__(self, name, value):
        if name in self.fields:
            raise AttributeError("Use item[{!r}] = {!r} to set field value".format(name, value))
        super(MutableMapping, self).__setattr__(name, value)

    def get_namespaces(self, assigned_only=True):
        """
        Get all namespaces of the elements and its attributes

        Parameters
        ----------
        assigned_only : bool
            Whether to return namespaces of assigned components only

        Returns
        -------
        set of (str or None, str or None)
            Set of pairs (namespace_prefix, namespace_uri)
        """
        namespaces = set()
        for elem in self.elements.values():
            if not assigned_only or elem.assigned:
                namespaces.update(elem.get_namespaces(assigned_only))
        return namespaces


ExtendableItem = FeedItem  # Backward compatibility



# -*- coding: utf-8 -*-

import pytest
import re
from itertools import product, chain, combinations

from scrapy_rss.meta import BaseNSComponent, NSComponentName
from scrapy_rss.exceptions import NoNamespaceURIError


names = ['name', 'sub_name', 'имя']
empty_ns_prefixes = [None, '']
ns_prefixes = ['ns_prefix', 'pre_prefix', 'префикс']
empty_ns_uris = [None, '']
ns_uris = ['id1', 'http://path.to/somewhere']


@pytest.mark.parametrize("ns_prefix,sec_ns_prefix,ns_uri",
                         chain(product(empty_ns_prefixes, ns_prefixes, ns_uris),
                               product(empty_ns_prefixes, empty_ns_prefixes, ns_uris)))
def test_ns_prefix_property_setter_v1(ns_prefix, sec_ns_prefix, ns_uri):
    c = BaseNSComponent(ns_prefix=ns_prefix, ns_uri=ns_uri)
    c.ns_prefix = sec_ns_prefix
    assert c.ns_prefix == sec_ns_prefix


@pytest.mark.parametrize("ns_prefix,sec_ns_prefix,ns_uri",
                         ((ns_prefix, sec_ns_prefix, ns_uri)
                          for ns_prefix, sec_ns_prefix in combinations(ns_prefixes, 2)
                          for ns_uri in  ns_uris))
def test_ns_prefix_property_setter_v2(ns_prefix, sec_ns_prefix, ns_uri):
    c = BaseNSComponent(ns_prefix=ns_prefix, ns_uri=ns_uri)
    with pytest.raises(ValueError, match=r"already non-empty"):
        c.ns_prefix = sec_ns_prefix


@pytest.mark.parametrize("ns_prefix,sec_ns_prefix,ns_uri",
                         product(empty_ns_prefixes, ns_prefixes, empty_ns_uris))
def test_ns_prefix_property_setter_v3(ns_prefix, sec_ns_prefix, ns_uri):
    c = BaseNSComponent(ns_prefix=ns_prefix, ns_uri=ns_uri)
    with pytest.raises(NoNamespaceURIError, match=r"no namespace URI"):
        c.ns_prefix = sec_ns_prefix


@pytest.mark.parametrize("ns_prefix,ns_uri,sec_ns_uri",
                         chain(product(empty_ns_prefixes, empty_ns_uris, ns_uris),
                               product(empty_ns_prefixes, empty_ns_uris, empty_ns_uris)))
def test_ns_uri_property_setter_v1(ns_prefix, ns_uri, sec_ns_uri):
    c = BaseNSComponent(ns_prefix=ns_prefix, ns_uri=ns_uri)
    c.ns_uri = sec_ns_uri
    assert c.ns_uri == sec_ns_uri


@pytest.mark.parametrize("ns_prefix,ns_uri,sec_ns_uri",
                         ((ns_prefix, ns_uri, sec_ns_uri)
                          for ns_prefix in ns_prefixes
                          for ns_uri, sec_ns_uri in  combinations(ns_uris, 2)))
def test_ns_uri_property_setter_v2(ns_prefix, ns_uri, sec_ns_uri):
    c = BaseNSComponent(ns_prefix=ns_prefix, ns_uri=ns_uri)
    with pytest.raises(ValueError, match=r"already non-empty"):
        c.ns_uri = sec_ns_uri




@pytest.mark.parametrize("name,ns_prefix,ns_uri", product(names, empty_ns_prefixes, empty_ns_uris))
def test_init_name_only(name, ns_prefix, ns_uri):
    NSComponentName(name)
    NSComponentName(name, ns_prefix)
    NSComponentName(name, ns_prefix=ns_prefix)
    NSComponentName(name, ns_uri=ns_uri)
    NSComponentName(name, ns_prefix, ns_uri)
    NSComponentName(name, ns_prefix=ns_prefix)
    NSComponentName(name, ns_prefix=ns_prefix, ns_uri=ns_uri)


@pytest.mark.parametrize("name,ns_prefix,sec_ns_prefix,ns_uri",
                         product(names, empty_ns_prefixes, ns_prefixes, empty_ns_uris))
def test_init_compound_name_only(name, ns_prefix, sec_ns_prefix, ns_uri):
    compound_name = sec_ns_prefix + '__' + name
    with pytest.raises(NoNamespaceURIError, match=re.escape(sec_ns_prefix)):
        NSComponentName(compound_name)
    with pytest.raises(NoNamespaceURIError, match=re.escape(sec_ns_prefix)):
        NSComponentName(compound_name, ns_prefix)
    with pytest.raises(NoNamespaceURIError, match=re.escape(sec_ns_prefix)):
        NSComponentName(compound_name, ns_prefix=ns_prefix)
    with pytest.raises(NoNamespaceURIError, match=re.escape(sec_ns_prefix)):
        NSComponentName(compound_name, ns_uri=ns_uri)
    with pytest.raises(NoNamespaceURIError, match=re.escape(sec_ns_prefix)):
        NSComponentName(compound_name, ns_prefix, ns_uri)
    with pytest.raises(NoNamespaceURIError, match=re.escape(sec_ns_prefix)):
        NSComponentName(compound_name, ns_prefix=ns_prefix)
    with pytest.raises(NoNamespaceURIError, match=re.escape(sec_ns_prefix)):
        NSComponentName(compound_name, ns_prefix=ns_prefix, ns_uri=ns_uri)


@pytest.mark.parametrize("name,ns_prefix,ns_uri", product(names, ns_prefixes, empty_ns_uris))
def test_init_name_and_prefix_only(name, ns_prefix, ns_uri):
    with pytest.raises(NoNamespaceURIError, match=re.escape(ns_prefix)):
        NSComponentName(name, ns_prefix)
    with pytest.raises(NoNamespaceURIError, match=re.escape(ns_prefix)):
        NSComponentName(name, ns_prefix=ns_prefix)
    with pytest.raises(NoNamespaceURIError, match=re.escape(ns_prefix)):
        NSComponentName(name, ns_prefix, ns_uri)
    with pytest.raises(NoNamespaceURIError, match=re.escape(ns_prefix)):
        NSComponentName(name, ns_prefix, ns_uri=ns_uri)


@pytest.mark.parametrize("name,ns_prefix,sec_ns_prefix,ns_uri", product(names, ns_prefixes, ns_prefixes, empty_ns_uris))
def test_init_compound_name_and_prefix_only(name, ns_prefix, sec_ns_prefix, ns_uri):
    compound_name = sec_ns_prefix + '__' + name
    with pytest.raises(NoNamespaceURIError, match=re.escape(ns_prefix)):
        NSComponentName(compound_name, ns_prefix)
    with pytest.raises(NoNamespaceURIError, match=re.escape(ns_prefix)):
        NSComponentName(compound_name, ns_prefix=ns_prefix)
    with pytest.raises(NoNamespaceURIError, match=re.escape(ns_prefix)):
        NSComponentName(compound_name, ns_prefix, ns_uri)
    with pytest.raises(NoNamespaceURIError, match=re.escape(ns_prefix)):
        NSComponentName(compound_name, ns_prefix, ns_uri=ns_uri)



@pytest.mark.parametrize("name,ns_prefix,ns_uri", chain(product(names, empty_ns_prefixes, ns_uris),
                                                        product(names, ns_prefixes, ns_uris)))
def test_init_name_prefix_uri(name, ns_prefix, ns_uri):
    NSComponentName(name, ns_uri=ns_uri)
    NSComponentName(name, ns_prefix, ns_uri)
    NSComponentName(name, ns_prefix=ns_prefix, ns_uri=ns_uri)


@pytest.mark.parametrize("name,ns_prefix,sec_ns_prefix,ns_uri",
                         chain(product(names, empty_ns_prefixes, ns_prefixes, ns_uris),
                               product(names, ns_prefixes, ns_prefixes, ns_uris)))
def test_init_compound_name_prefix_uri(name, ns_prefix, sec_ns_prefix, ns_uri):
    compound_name = sec_ns_prefix + '__' + name
    NSComponentName(compound_name, ns_uri=ns_uri)
    NSComponentName(compound_name, ns_prefix, ns_uri)
    NSComponentName(compound_name, ns_prefix=ns_prefix, ns_uri=ns_uri)


@pytest.mark.parametrize("name,ns_prefix,ns_uri", chain(product(names, empty_ns_prefixes, empty_ns_uris),
                                                        product(names, empty_ns_prefixes, ns_uris),
                                                        product(names, ns_prefixes, ns_uris)))
def test_name_property(name, ns_prefix, ns_uri):
    n = NSComponentName(name, ns_prefix, ns_uri)
    assert n.name == name
    n = NSComponentName(name, ns_prefix=ns_prefix, ns_uri=ns_uri)
    assert n.name == name

@pytest.mark.parametrize("name,ns_prefix,sec_ns_prefix,ns_uri",
                         chain(product(names, empty_ns_prefixes, ns_prefixes, ns_uris),
                               product(names, ns_prefixes, ns_prefixes, ns_uris)))
def test_name_property_with_compound_name(name, ns_prefix, sec_ns_prefix, ns_uri):
    compound_name = sec_ns_prefix + '__' + name
    n = NSComponentName(compound_name, ns_prefix, ns_uri)
    assert n.name == name
    n = NSComponentName(compound_name, ns_prefix=ns_prefix, ns_uri=ns_uri)
    assert n.name == name


@pytest.mark.parametrize("name,ns_prefix,ns_uri", chain(product(names, empty_ns_prefixes, empty_ns_uris),
                                                        product(names, empty_ns_prefixes, ns_uris),
                                                        product(names, ns_prefixes, ns_uris)))
def test_xml_name_property(name, ns_prefix, ns_uri):
    n = NSComponentName(name, ns_prefix, ns_uri)
    assert n.xml_name == (ns_uri or '', name)
    n = NSComponentName(name, ns_prefix=ns_prefix, ns_uri=ns_uri)
    assert n.xml_name == (ns_uri or '', name)


@pytest.mark.parametrize("name,ns_prefix,sec_ns_prefix,ns_uri",
                         chain(product(names, empty_ns_prefixes, ns_prefixes, ns_uris),
                               product(names, ns_prefixes, ns_prefixes, ns_uris)))
def test_xml_name_property_with_compound_name(name, ns_prefix, sec_ns_prefix, ns_uri):
    compound_name = sec_ns_prefix + '__' + name
    n = NSComponentName(compound_name, ns_prefix, ns_uri)
    assert n.xml_name == (ns_uri or '', name)
    n = NSComponentName(compound_name, ns_prefix=ns_prefix, ns_uri=ns_uri)
    assert n.xml_name == (ns_uri or '', name)


@pytest.mark.parametrize("name,ns_prefix,ns_uri", chain(product(names, empty_ns_prefixes, empty_ns_uris),
                                                        product(names, empty_ns_prefixes, ns_uris),
                                                        product(names, ns_prefixes, ns_uris)))
def test_pub_name_property(name, ns_prefix, ns_uri):
    n = NSComponentName(name, ns_prefix, ns_uri)
    assert n.pub_name == name
    n = NSComponentName(name, ns_prefix=ns_prefix, ns_uri=ns_uri)
    assert n.pub_name == name


@pytest.mark.parametrize("name,ns_prefix,sec_ns_prefix,ns_uri",
                         chain(product(names, empty_ns_prefixes, ns_prefixes, ns_uris),
                               product(names, ns_prefixes, ns_prefixes, ns_uris)))
def test_pub_name_property_with_compound_name(name, ns_prefix, sec_ns_prefix, ns_uri):
    compound_name = sec_ns_prefix + '__' + name
    n = NSComponentName(compound_name, ns_prefix, ns_uri)
    assert n.pub_name == compound_name
    n = NSComponentName(compound_name, ns_prefix=ns_prefix, ns_uri=ns_uri)
    assert n.pub_name == compound_name


@pytest.mark.parametrize("name,ns_prefix,ns_uri", chain(product(names, empty_ns_prefixes, empty_ns_uris),
                                                        product(names, empty_ns_prefixes, ns_uris),
                                                        product(names, ns_prefixes, ns_uris)))
def test_str(name, ns_prefix, ns_uri):
    n = NSComponentName(name, ns_prefix, ns_uri)
    assert str(n) == name
    n = NSComponentName(name, ns_prefix=ns_prefix, ns_uri=ns_uri)
    assert str(n) == name


@pytest.mark.parametrize("name,ns_prefix,sec_ns_prefix,ns_uri",
                         chain(product(names, empty_ns_prefixes, ns_prefixes, ns_uris),
                               product(names, ns_prefixes, ns_prefixes, ns_uris)))
def test_str_with_compound_name(name, ns_prefix, sec_ns_prefix, ns_uri):
    compound_name = sec_ns_prefix + '__' + name
    n = NSComponentName(compound_name, ns_prefix, ns_uri)
    assert str(n) == compound_name
    n = NSComponentName(compound_name, ns_prefix=ns_prefix, ns_uri=ns_uri)
    assert str(n) == compound_name


@pytest.mark.parametrize("name,ns_prefix,ns_uri", chain(product(names, empty_ns_prefixes, empty_ns_uris),
                                                        product(names, empty_ns_prefixes, ns_uris),
                                                        product(names, ns_prefixes, ns_uris)))
def test_priv_name_property(name, ns_prefix, ns_uri):
    n = NSComponentName(name, ns_prefix, ns_uri)
    assert n.priv_name == '__' + name
    n = NSComponentName(name, ns_prefix=ns_prefix, ns_uri=ns_uri)
    assert n.priv_name == '__' + name


@pytest.mark.parametrize("name,ns_prefix,sec_ns_prefix,ns_uri",
                         chain(product(names, empty_ns_prefixes, ns_prefixes, ns_uris),
                               product(names, ns_prefixes, ns_prefixes, ns_uris)))
def test_priv_name_property_with_compound_name(name, ns_prefix, sec_ns_prefix, ns_uri):
    compound_name = sec_ns_prefix + '__' + name
    n = NSComponentName(compound_name, ns_prefix, ns_uri)
    assert n.priv_name == '__' + compound_name
    n = NSComponentName(compound_name, ns_prefix=ns_prefix, ns_uri=ns_uri)
    assert n.priv_name == '__' + compound_name


@pytest.mark.parametrize("name,ns_prefix,ns_uri", product(names, empty_ns_prefixes, empty_ns_uris))
def test_implemented_comparison_v1(name, ns_prefix, ns_uri):
    n1 = NSComponentName(name, ns_prefix, ns_uri)
    n2 = NSComponentName(name, ns_uri=ns_uri, ns_prefix=ns_prefix)
    n3 = NSComponentName(name + '9', ns_prefix, ns_uri)
    assert n1 == n2
    assert n2 == n1
    assert n1 != n3
    assert n3 != n1
    assert n2 != n3
    assert n3 != n2


@pytest.mark.parametrize("name,ns_prefix,ns_uri", chain(product(names, empty_ns_prefixes, ns_uris),
                                                        product(names, ns_prefixes, ns_uris)))
def test_comparison_v2(name, ns_prefix, ns_uri):
    n1 = NSComponentName(name, ns_prefix, ns_uri)
    n2 = NSComponentName(name, ns_uri=ns_uri, ns_prefix=ns_prefix)
    n3 = NSComponentName(name + '9', ns_prefix, ns_uri)
    n4 = NSComponentName(name, ns_prefix, ns_uri + '2')
    assert n1 == n2
    assert n2 == n1
    assert n1 != n3
    assert n3 != n1
    assert n2 != n3
    assert n3 != n2
    assert n1 != n4
    assert n4 != n1
    assert n2 != n4
    assert n4 != n2
    assert n3 != n4
    assert n4 != n3


@pytest.mark.parametrize("name,ns_prefix,sec_ns_prefix,ns_uri",
                         chain(product(names, empty_ns_prefixes, ns_prefixes, ns_uris),
                               product(names, ns_prefixes, ns_prefixes, ns_uris)))
def test_comparison_with_compound_name(name, ns_prefix, sec_ns_prefix, ns_uri):
    compound_name = sec_ns_prefix + '__' + name
    n1 = NSComponentName(compound_name, ns_prefix, ns_uri)
    n2 = NSComponentName(compound_name, ns_uri=ns_uri, ns_prefix=ns_prefix)
    n3 = NSComponentName(compound_name + '9', ns_prefix, ns_uri)
    n4 = NSComponentName(compound_name, ns_prefix, ns_uri + '2')
    assert n1 == n2
    assert n2 == n1
    assert n1 != n3
    assert n3 != n1
    assert n2 != n3
    assert n3 != n2
    assert n1 != n4
    assert n4 != n1
    assert n2 != n4
    assert n4 != n2
    assert n3 != n4
    assert n4 != n3


@pytest.mark.parametrize("name,ns_prefix,ns_uri", chain(product(names, empty_ns_prefixes, empty_ns_uris),
                                                        product(names, empty_ns_prefixes, ns_uris),
                                                        product(names, ns_prefixes, ns_uris)))
def test_unsupported_comparison(name, ns_prefix, ns_uri):
    n = NSComponentName(name, ns_prefix, ns_uri)
    with pytest.raises(NotImplementedError):
        n == (ns_uri, name)
    with pytest.raises(NotImplementedError):
        (ns_uri, name) == n
    with pytest.raises(NotImplementedError):
        n.pub_name == n
    with pytest.raises(NotImplementedError):
        n == n.pub_name


@pytest.mark.parametrize("name,ns_prefix,sec_ns_prefix,ns_uri",
                         chain(product(names, empty_ns_prefixes, ns_prefixes, ns_uris),
                               product(names, ns_prefixes, ns_prefixes, ns_uris)))
def test_unsupported_comparison_with_compound_name(name, ns_prefix, sec_ns_prefix, ns_uri):
    compound_name = sec_ns_prefix + '__' + name
    n = NSComponentName(compound_name, ns_prefix, ns_uri)
    with pytest.raises(NotImplementedError):
        n == (ns_uri, name)
    with pytest.raises(NotImplementedError):
        (ns_uri, name) == n
    with pytest.raises(NotImplementedError):
        n.pub_name == n
    with pytest.raises(NotImplementedError):
        n == n.pub_name


@pytest.mark.parametrize("name,ns_prefix,ns_uri", chain(product(names, empty_ns_prefixes, empty_ns_uris),
                                                        product(names, empty_ns_prefixes, ns_uris),
                                                        product(names, ns_prefixes, ns_uris)))
def test_repr(name, ns_prefix, ns_uri):
    n = NSComponentName(name, ns_prefix, ns_uri)
    assert repr(n) == "NSComponentName(name={!r}, ns_prefix={!r}, ns_uri={!r})".format(name, ns_prefix or '', ns_uri or '')


@pytest.mark.parametrize("name,ns_prefix,sec_ns_prefix,ns_uri",
                         product(names, empty_ns_prefixes, ns_prefixes, ns_uris))
def test_repr_with_compound_name_v1(name, ns_prefix, sec_ns_prefix, ns_uri):
    compound_name = sec_ns_prefix + '__' + name
    n = NSComponentName(compound_name, ns_prefix, ns_uri)
    assert repr(n) == "NSComponentName(name={!r}, ns_prefix={!r}, ns_uri={!r})".format(name, sec_ns_prefix, ns_uri)



@pytest.mark.parametrize("name,ns_prefix,sec_ns_prefix,ns_uri",
                         product(names, ns_prefixes, ns_prefixes, ns_uris))
def test_repr_with_compound_name_v2(name, ns_prefix, sec_ns_prefix, ns_uri):
    compound_name = sec_ns_prefix + '__' + name
    n = NSComponentName(compound_name, ns_prefix, ns_uri)
    assert repr(n) == "NSComponentName(name={!r}, ns_prefix={!r}, ns_uri={!r})".format(name, ns_prefix, ns_uri)


@pytest.mark.parametrize("name,ns_prefix,ns_uri", product(names, empty_ns_prefixes, empty_ns_uris))
def test_get_namespaces_v1(name, ns_prefix, ns_uri):
    n = NSComponentName(name, ns_prefix, ns_uri)
    namespaces = n.get_namespaces()
    assert isinstance(namespaces, set)
    assert namespaces == set()


@pytest.mark.parametrize("name,ns_prefix,ns_uri", chain(product(names, empty_ns_prefixes, ns_uris),
                                                        product(names, ns_prefixes, ns_uris)))
def test_get_namespaces_v2(name, ns_prefix, ns_uri):
    n = NSComponentName(name, ns_prefix, ns_uri)
    assert n.get_namespaces() == {(ns_prefix or '', ns_uri)}


@pytest.mark.parametrize("name,ns_prefix,sec_ns_prefix,ns_uri",
                         product(names, empty_ns_prefixes, ns_prefixes, ns_uris))
def test_get_namespaces_with_compound_name_v1(name, ns_prefix, sec_ns_prefix, ns_uri):
    compound_name = sec_ns_prefix + '__' + name
    n = NSComponentName(compound_name, ns_prefix, ns_uri)
    assert n.get_namespaces() == {(sec_ns_prefix, ns_uri)}


@pytest.mark.parametrize("name,ns_prefix,sec_ns_prefix,ns_uri",
                         product(names, ns_prefixes, ns_prefixes, ns_uris))
def test_get_namespaces_with_compound_name_v2(name, ns_prefix, sec_ns_prefix, ns_uri):
    compound_name = sec_ns_prefix + '__' + name
    n = NSComponentName(compound_name, ns_prefix, ns_uri)
    assert n.get_namespaces() == {(ns_prefix, ns_uri)}


if __name__ == '__main__':
    pytest.main()



# -*- coding: utf-8 -*-

from itertools import chain
from collections import Counter

from packaging import version
from datetime import datetime
from dateutil.tz import tzlocal
import six
import scrapy
from scrapy.exporters import XmlItemExporter

from .items import RssItem
from .utils import format_rfc822
from .exceptions import *
from . import meta


class RssItemExporter(XmlItemExporter):
    def __init__(self, file, channel_title, channel_link, channel_description,
                 namespaces=None, item_cls=None,
                 language=None, copyright=None, managing_editor=None, webmaster=None,
                 pubdate=None, last_build_date=None, category=None, generator=None,
                 docs=None, ttl=None,
                 **kwargs):
        """
        RSS parameters semantics: https://validator.w3.org/feed/docs/rss2.html

        Parameters
        ----------
        namespaces : {str or None : str} or tuple of (str or None, str) or list of (str or None, str) or None
            predefined XML namespaces {prefix: URI, ...} or [(prefix, URI), ...]
        item_cls : type
            main class of RSS items (default: RssItem)

        Item fields that corresponds to RSS element with attributes or sub-elements,
        must be a dictionary-like such as
        {"attribute-name": "attribute-value"} (with special key "content" if RSS element must have content)
        or {"sub-element-name": "sub-element-content"}

        For example, the dictionary for <guid> element:
            >>> {'isPermalink': 'False', 'content': '0123456789abcdef'}
            that converts into
            <guid isPermalink="False">0123456789abcdef</guid>
        """
        kwargs['root_element'] = 'rss'
        kwargs['item_element'] = 'item'
        super(RssItemExporter, self).__init__(file, **kwargs)

        self.channel_element = 'channel'

        self.channel_title = channel_title
        self.channel_link = channel_link
        self.channel_description = channel_description

        self.channel_language = language
        self.channel_copyright = copyright
        if managing_editor and '@' not in managing_editor:
            raise ValueError('managing_editor field must contain at least e-mail. Passed: {}'.format(managing_editor))
        self.channel_managing_editor = managing_editor
        if webmaster and '@' not in webmaster:
            raise ValueError('webmaster field must contain at least e-mail. Passed: {}'.format(webmaster))
        self.channel_webmaster = webmaster
        self.channel_pubdate = pubdate
        self.channel_last_build_date = last_build_date if last_build_date \
            else datetime.today().replace(tzinfo=tzlocal())
        self.channel_category = ([category] if category and isinstance(category, six.string_types)
                                 and not isinstance(category, (list, set, tuple))
                                 else category)
        self.channel_generator = generator if generator is not None else 'Scrapy {}'.format(scrapy.__version__)
        self.channel_docs = docs
        self.channel_ttl = ttl

        if not item_cls:
            item_cls = RssItem
        elif not issubclass(item_cls, RssItem):
            raise ValueError('Item class must be RssItem class or its subclass')
        self._item_cls = item_cls

        if not namespaces:
            namespaces = {}
        elif isinstance(namespaces, (list, tuple)):
            namespaces = dict(namespaces)
        namespaces_iter = namespaces.items() if isinstance(namespaces, dict) else namespaces
        item_cls_namespaces = item_cls().get_namespaces(False)
        self._namespaces = {}
        skipped_ns_prefixes = set()
        for ns_prefix, ns_uri in chain(namespaces_iter, item_cls_namespaces):
            if ns_prefix in skipped_ns_prefixes:
                continue
            if ns_prefix in self._namespaces and ns_uri != self._namespaces[ns_prefix]:
                self._namespaces.pop(ns_prefix)
                skipped_ns_prefixes.add(ns_prefix)
            else:
                self._namespaces[ns_prefix] = ns_uri

    if version.parse(scrapy.__version__) < version.parse('1.4.0'):
        def _export_xml_field(self, name, serialized_value, depth):
            return super(RssItemExporter, self)._export_xml_field(name, serialized_value)

    def start_exporting(self):
        self.xg.startDocument()
        for ns_prefix, ns_uri in self._namespaces.items():
            self.xg.startPrefixMapping(ns_prefix, ns_uri)
        root_attrs = {(None, 'version'): '2.0'}
        self.xg.startElementNS((None, self.root_element), self.root_element, root_attrs)
        self.xg.startElement(self.channel_element, {})

        self._export_xml_field('title', self.channel_title, 1)
        self._export_xml_field('link', self.channel_link, 1)
        self._export_xml_field('description', self.channel_description, 1)
        if self.channel_language:
            self._export_xml_field('language', self.channel_language, 1)
        if self.channel_copyright:
            self._export_xml_field('copyright', self.channel_copyright, 1)
        if self.channel_managing_editor:
            self._export_xml_field('managingEditor', self.channel_managing_editor, 1)
        if self.channel_webmaster:
            self._export_xml_field('webMaster', self.channel_webmaster, 1)
        if self.channel_pubdate:
            self._export_xml_field('pubdate',
                                   format_rfc822(self.channel_pubdate)
                                   if isinstance(self.channel_pubdate, datetime)
                                   else self.channel_pubdate, 1)
        self._export_xml_field('lastBuildDate',
                               format_rfc822(self.channel_last_build_date)
                               if isinstance(self.channel_last_build_date, datetime)
                               else self.channel_last_build_date, 1)
        if self.channel_category:
            for category in self.channel_category:
                self._export_xml_field('category', category, 1)
        if self.channel_generator:
            self._export_xml_field('generator', self.channel_generator, 1)
        if self.channel_docs:
            self._export_xml_field('docs', self.channel_docs, 1)
        if self.channel_ttl:
            self._export_xml_field('ttl', self.channel_ttl, 1)

    def export_item(self, item):
        if not isinstance(item, RssItem) and not isinstance(getattr(item, 'rss', None), RssItem):
            raise InvalidRssItemError("Item must have 'rss' field of type 'RssItem'")
        if not isinstance(item, RssItem):
            item = item.rss

        item_namespaces = set()
        if item.__class__ is not self._item_cls:
            item_namespaces = item.get_namespaces()
            item_namespaces -= set(self._namespaces.items())
            ns_prefix_count = Counter(ns_prefix for ns_prefix, _ in item_namespaces)
            item_namespaces = {ns for ns in item_namespaces if ns_prefix_count[ns[0]] == 1}
        item_namespaces = dict(item_namespaces)

        for elem_ns_prefix, elem_ns_uri in item_namespaces.items():
            self.xg.startPrefixMapping(elem_ns_prefix, elem_ns_uri)
        self.xg.startElementNS((None, self.item_element), self.item_element, {})

        for elem_name, elem_descr in item.elements.items():
            elem_value = getattr(item, str(elem_name))
            if elem_value.assigned:
                elem_values = elem_value if isinstance(elem_value, meta.MultipleElements) else (elem_value,)
                for elem_value in elem_values:
                    if not elem_value.is_valid():
                        raise InvalidRssItemAttributesError(elem_name,
                                                            list(elem_value.required_attrs),
                                                            elem_value.content_arg)

                    attrs = elem_value.serialize()
                    content = attrs.pop(elem_descr.content_arg.xml_name, None) if elem_descr.content_arg else None
                    undeclared_elem_namespaces = {ns_prefix: ns_uri
                                                  for ns_prefix, ns_uri in elem_descr.get_namespaces()
                                                  if (ns_prefix not in self._namespaces 
                                                      or self._namespaces[ns_prefix] != ns_uri)
                                                  and (ns_prefix not in item_namespaces 
                                                       or item_namespaces[ns_prefix] != ns_uri)}
                    if elem_name.ns_prefix:
                        elem_qname = '{}:{}'.format(elem_name.ns_prefix, elem_name.name)
                    else:
                        elem_qname = elem_name.name
                    for ns_prefix, ns_uri in undeclared_elem_namespaces.items():
                        self.xg.startPrefixMapping(ns_prefix, ns_uri)
                    self.xg.startElementNS(elem_name.xml_name, elem_qname, attrs)
                    if content:
                        self.xg.characters(content)
                    self.xg.endElementNS(elem_name.xml_name, elem_qname)
                    for ns_prefix, ns_uri in undeclared_elem_namespaces.items():
                        self.xg.endPrefixMapping(ns_prefix)

        self.xg.endElementNS((None, self.item_element), self.item_element)
        for elem_ns_prefix in item_namespaces:
            self.xg.endPrefixMapping(elem_ns_prefix)

    def finish_exporting(self):
        self.xg.endElement(self.channel_element)
        self.xg.endElementNS((None, self.root_element), self.root_element)
        for ns_prefix, ns_uri in self._namespaces.items():
            self.xg.startPrefixMapping(ns_prefix, ns_uri)
        self.xg.endDocument()



# -*- coding: utf-8 -*-

from copy import deepcopy
import re
import six

from .attribute import ItemElementAttribute
from .nscomponent import BaseNSComponent, NSComponentName


class ItemElementMeta(type):
    def __new__(mcs, cls_name, cls_bases, cls_attrs):
        elem_attrs = {NSComponentName(attr_name, ns_prefix=attr_descr.ns_prefix, ns_uri=attr_descr.ns_uri):
                      attr_descr for attr_name, attr_descr in cls_attrs.items()
                      if isinstance(attr_descr, ItemElementAttribute)}
        for attr_name, attr in elem_attrs.items():
            if not attr.ns_prefix:
                attr.ns_prefix = attr_name.ns_prefix
        if sum(attr.is_content for attr in elem_attrs.values()) > 1:
            raise ValueError("More than one attributes that's interpreted as content in the element '{}' specification"
                             .format(cls_name))
        cls_attrs['_attrs'] = elem_attrs

        cls_attrs.update({attr_name.priv_name: attr_descr
                          for attr_name, attr_descr in elem_attrs.items()})
        cls_attrs.update({attr_name.pub_name:
                          property(mcs.build_attr_getter(attr_name),
                                   mcs.build_attr_setter(attr_name))
                          for attr_name, attr_descr in elem_attrs.items()})

        return super(ItemElementMeta, mcs).__new__(mcs, cls_name, cls_bases, cls_attrs)

    def __init__(cls, cls_name, cls_bases, cls_attrs):
        cls._assigned = False
        cls.assigned = property(lambda self: self._assigned)
        cls.attrs = property(lambda self: self._attrs) # Attributes

        cls._required_attrs = {attr_name
                               for attr_name, attr_descr in cls._attrs.items()
                               if attr_descr.required and not attr_descr.is_content}
        cls.required_attrs = property(lambda self: self._required_attrs)

        cls._content_arg = None
        for attr_name, attr_descr in cls._attrs.items():
            if attr_descr.is_content:
                cls._content_arg = attr_name
                break
        cls.content_arg = property(lambda self: self._content_arg)

        cls.serialize = lambda self: {attr_name.xml_name: attr.serializer(attr.value)
                                      for attr_name in self.attrs
                                      for attr in (getattr(self, attr_name.priv_name),)
                                      if attr.value is not None}

        super(ItemElementMeta, cls).__init__(cls_name, cls_bases, cls_attrs)

    @staticmethod
    def build_attr_getter(name):
        """
        Build attribute getter

        Parameters
        ----------
        name : NSComponentName
            an attribute name
        """
        return lambda self: getattr(self, name.priv_name).value

    @staticmethod
    def build_attr_setter(name):
        """
        Build attribute setter

        Parameters
        ----------
        name : NSComponentName
            an attribute name
        """
        def setter(self, value):
            attr = getattr(self, name.priv_name)
            attr.value = value
            self._assigned = True
        return setter


@six.add_metaclass(ItemElementMeta)
class ItemElement(BaseNSComponent):
    """
    Base class for elements

    Attributes
    ----------
    attrs : {NSComponentName : ItemElementAttribute}
        All attributes of the element
    required_attrs : set of NSComponentName
        Required element attributes
    content_arg : NSComponentName or None
        Name of an attribute that's interpreted as the element content
    assigned : bool
        Whether a non-None value is assigned to any attribute of the element

    Methods
    -------
    serialize() : { (str or None, str) : str }
        Convert values of element attributes to a strings.
        The dictionary key is a tuple (namespace_uri, attribute_name) for SAX handlers
    """

    def __init__(self, *args, **kwargs):
        if not self.content_arg and args:
            raise ValueError("Element of type '{}' does not support unnamed arguments (no content)"
                             .format(self.__class__.__name__))
        if len(args) > 1:
            raise ValueError("Constructor of class '{}' supports only single unnamed argument "
                             "that is interpreted as content of element".format(self.__class__.__name__))

        if args and str(self.content_arg) not in kwargs:
            kwargs[str(self.content_arg)] = args[0]

        new_attrs = {}
        for attr_name in self._attrs:
            new_attr = deepcopy(getattr(self, attr_name.priv_name))
            setattr(self, attr_name.priv_name, new_attr)
            new_attrs[attr_name] = new_attr
        self._attrs = new_attrs

        for attr_name in self._attrs:
            attr_name = str(attr_name)
            if attr_name in kwargs:
                setattr(self, attr_name, kwargs[attr_name])
                del kwargs[attr_name]

        try:
            super(ItemElement, self).__init__(**kwargs)
        except TypeError:
            raise ValueError("Passed arguments {}. "
                             "But constructor of class '{}' supports only the next named arguments: {}"
                             .format(list(kwargs.keys()), self.__class__.__name__,
                                     [str(a) for a in self.attrs]))

    def __repr__(self):
        s_match = re.match(r'^[^(]+\((.*?)\)$', super(ItemElement, self).__repr__())
        s_repr = s_match.group(1) if s_match else ''
        attrs_repr = ", ".join("{}={!r}".format(attr_name, attr)
                               for attr_name, attr in self.attrs.items())
        return "{}({})".format(self.__class__.__name__, ", ".join(filter(None, [attrs_repr, s_repr])))

    def is_valid(self):
        """
        Check if the element has valid attributes' values
        """
        return (not self._assigned
                or all(getattr(self, str(attr_name)) is not None for attr_name in self.required_attrs)
                and (not self.content_arg or not self.attrs[self.content_arg].required
                     or getattr(self, str(self.content_arg)) is not None))

    def get_namespaces(self, assigned_only=True):
        namespaces = super(ItemElement, self).get_namespaces()
        for attr in self.attrs.values():
            if not assigned_only or attr.value is not None:
                namespaces.update(attr.get_namespaces(assigned_only))
        return namespaces


class MultipleElements(ItemElement):
    """
    Represents elements of the same base class
    """

    def __init__(self, base_element_cls, **kwargs):
        if not isinstance(base_element_cls, ItemElementMeta):
            raise TypeError("Invalid type of elements class: {}".format(base_element_cls))
        self.base_element_cls = base_element_cls
        super(MultipleElements, self).__init__(**kwargs)
        self.elements = []
        self._kwargs = kwargs
        self._content_arg = base_element_cls._content_arg

        def serializer():
            raise NotImplementedError('Class MultipleElements does not support serialization')
        self.serialize = serializer

    def _check_value(self, value):
        if not self._kwargs.get("ns_prefix") and self.ns_prefix:
            self._kwargs["ns_prefix"] = self.ns_prefix
        if isinstance(value, self.base_element_cls):
            if self._kwargs.get("ns_uri") and not value.ns_uri:
                value.ns_uri = self._kwargs["ns_uri"]
            if not value.ns_prefix and self.ns_prefix:
                value.ns_prefix = self.ns_prefix
            return value
        if isinstance(value, dict):
            kwargs = self._kwargs.copy()
            kwargs.update(value)
            elem = self.base_element_cls(**kwargs)
            return elem
        return self.base_element_cls(value, **self._kwargs)

    def append(self, elem):
        """
        Append new element

        Parameters
        ----------
        elem : ItemElement or { str : Any }
            New element as element instance or dictionary of attributes' values
        """
        self.elements.append(self._check_value(elem))
        self._assigned = True

    def extend(self, iterable):
        """
        Add multiple elements

        Parameters
        ----------
        iterable : iterable of (ItemElement or { str : Any })
            Iterable of elements as element instances or dictionaries of attributes' values
        """
        for elem in iterable:
            self.append(elem)

    def add(self, value):
        """
        Add element(s) by value

        Parameters
        ----------
        value : ItemElement or { str : Any } or iterable of (ItemElement or { str : Any })
            New element(s) as element instance(s) or dictionary(-ies) of attributes' values
        """
        if isinstance(value, list):
            self.extend(value)
        else:
            self.append(value)

    def clear(self):
        """
        Remove all elements
        """
        del self.elements[:]
        self._assigned = False

    def pop(self, index=-1):
        """
        Remove single element

        Parameters
        ----------
        index : int
            Element index (last by default)
        """
        elem = self.elements.pop(index)
        if not self.elements:
            self._assigned = False
        return elem

    def __delitem__(self, index):
        self.elements.__delitem__(index)
        if not self.elements:
            self._assigned = False

    def __getitem__(self, index):
        return self.elements[index]

    def __iter__(self):
        return iter(self.elements)

    def __len__(self):
        return len(self.elements)

    def __setitem__(self, index, elem):
        if not isinstance(elem, self.base_element_cls):
            raise TypeError("Elements must have type '{}' or descendant type, not '{}'"
                            .format(self.base_element_cls, elem.__class__))
        self.elements[index] = elem

    def __getattr__(self, name):
        if name == 'base_element_cls':
            raise AttributeError("'{}' object has no attribute '{}'".format(self.__class__.__name__, name))
        if name not in {attr.pub_name for attr in self.base_element_cls._attrs}:
            raise AttributeError("Elements of type '{}' does not have '{}' attribute"
                                 .format(self.base_element_cls, name))
        if not self.elements:
            raise AttributeError("Instances of '{}' have not been assigned"
                                 .format(self.base_element_cls.__name__))
        if len(self.elements) > 1:
            raise AttributeError("Cannot get attribute: more than one elements have been assigned. "
                                 "Choose element and get its' attribute.")
        return getattr(self.elements[0], name)

    def __setattr__(self, name, value):
        if name != 'base_element_cls' and name in {attr.pub_name for attr in self.base_element_cls._attrs}:
            if len(self.elements) != 1:
                raise AttributeError("Cannot set attribute: {} elements have been assigned. "
                                     "Choose element and set its' attribute.".format(len(self.elements)))
            setattr(self.elements[0], name, value)
        else:
            return super(MultipleElements, self).__setattr__(name, value)

    def __repr__(self):
        s_match = re.match(r'^[^(]+\((.*?)\)$', super(MultipleElements, self).__repr__())
        s_repr = s_match.group(1) if s_match else ''
        base_cls_repr = "base_element_cls={!r}".format(self.base_element_cls)
        return "{}({})".format(self.__class__.__name__, ", ".join(filter(None, [base_cls_repr, s_repr])))

    def get_namespaces(self, assigned_only=True):
        namespaces = super(MultipleElements, self).get_namespaces()
        for elem in self.elements:
            namespaces.update(elem.get_namespaces(assigned_only))
        return namespaces



# -*- coding: utf-8 -*-

import re

from ..exceptions import NoNamespaceURIError


class BaseNSComponent(object):
    def __init__(self, ns_prefix=None, ns_uri=None):
        """
        Base class for elements, attributes and its names that can be namespaced

        Parameters
        ----------
        ns_prefix : str or None
            a namespace prefix
        ns_uri : str or None
            a namespace URI
        """
        if ns_prefix and not ns_uri:
            raise NoNamespaceURIError("No URI for prefix '{}'".format(ns_prefix))
        self._ns_prefix = ns_prefix or ''
        self._ns_uri = ns_uri or ''

    @property
    def ns_prefix(self):
        """
        Get the namespace prefix

        Returns
        -------
        str or None
            The namespace prefix
        """
        return self._ns_prefix

    @ns_prefix.setter
    def ns_prefix(self, ns_prefix):
        """
        Set the namespace prefix

        Parameters
        ----------
        ns_prefix : str or None
            A new namespace prefix

        Raises
        ------
        NoNamespaceURIError
            If component does not have namespace URI
        ValueError
            If the namespace prefix is already non-empty
        """
        if self._ns_prefix == ns_prefix:
            return
        if not self._ns_uri:
            raise NoNamespaceURIError("Namespace prefix cannot be set when no namespace URI")
        if self._ns_prefix:
            raise ValueError("Namespace prefix is already non-empty")
        self._ns_prefix = ns_prefix

    @property
    def ns_uri(self):
        """
        Get the namespace URI

        Returns
        -------
        str or None
            The namespace URI
        """
        return self._ns_uri

    @ns_uri.setter
    def ns_uri(self, ns_uri):
        """
        Set the namespace URI

        Parameters
        ----------
        ns_uri : str or None
            A new namespace URI

        Raises
        ------
        ValueError
            If the namespace URI is already non-empty
        """
        if self._ns_uri == ns_uri:
            return
        if self._ns_uri:
            raise ValueError("Namespace URI is already non-empty")
        self._ns_uri = ns_uri

    def __repr__(self):
        return "{}(ns_prefix={!r}, ns_uri={!r})"\
            .format(self.__class__.__name__, self._ns_prefix, self._ns_uri)

    def get_namespaces(self, assigned_only=True):
        """
        Get all namespaces of the component and its children

        Parameters
        ----------
        assigned_only : bool
            whether return namespaces of assigned children only

        Returns
        -------
        set of (str or None, str or None)
            Set of pairs (namespace_prefix, namespace_uri)
        """
        if self._ns_uri:
            return {(self._ns_prefix, self._ns_uri)}
        return set()


class NSComponentName(BaseNSComponent):
    def __init__(self, name, ns_prefix=None, ns_uri=None):
        """
        Component' name wrapper

        Parameters
        ----------
        name : str
            the component name that can optionally contain a namespace prefix
            using delimiter __ (double underscores) such as ns__name
        ns_prefix : str or None
            a namespace prefix
        ns_uri : str or None
            a namespace URI
        """
        if '__' in name:
            secondary_ns_prefix, name = name.split('__', 1)
        else: 
            secondary_ns_prefix = None
        super(NSComponentName, self).__init__(ns_prefix=ns_prefix or secondary_ns_prefix,
                                              ns_uri=ns_uri)
        self._name = name
        if secondary_ns_prefix:
            self._public_fullname = '{}__{}'.format(secondary_ns_prefix, name)
        else:
            self._public_fullname = name
        self._private_fullname = '__{}'.format(self._public_fullname) 

    @property
    def name(self):
        """
        Get main name of the component without namespace information

        Returns
        -------
        str
            main component name
        """
        return self._name

    @property
    def xml_name(self):
        """
        Get tuple **(namespace_uri, name)** for XML handlers

        Returns
        -------
        (str or None, str)
            component name in the namespaced SAX format
        """
        return self._ns_uri, self._name

    @property
    def pub_name(self):
        """
        Get component name with namespace URI in the Python public notation format
        such as **uri__name** or **name** if no namespace

        Returns
        -------
        str
            public component name
        """
        return self._public_fullname

    @property
    def priv_name(self):
        """
        Get component name with namespace URI in the Python pseudo-private notation format
        such as **__uri__name** or **__name** if no namespace

        Returns
        -------
        str
            private component name
        """
        return self._private_fullname

    def __str__(self):
        return self._public_fullname

    def __key(self):
        return self._ns_uri, self._name

    def __hash__(self):
        return hash(self.__key())

    def __eq__(self, other):
        if isinstance(other, NSComponentName):
            return self.__key() == other.__key()
        raise NotImplementedError("Cannot compare instances of {} and {}".format(self.__class__, other.__class__))

    def __repr__(self):
        s_match = re.match(r'^[^(]+\((.*?)\)$', super(NSComponentName, self).__repr__())
        s_repr = ", " + s_match.group(1) if s_match else ''
        return "{}(name={!r}{})".format(self.__class__.__name__, self._name, s_repr)



# -*- coding: utf-8 -*-

import locale


def format_rfc822(date):
    """

    Parameters
    ----------
    date : datetime.datetime
        Datetime object

    Returns
    -------
    str
        Stringified datetime object according to RFC 822 standard
    """
    orig_locale = locale.getlocale(locale.LC_TIME)
    locale.setlocale(locale.LC_TIME, 'C')
    date = date.strftime('%a, %d %b %Y %H:%M:%S %z')
    locale.setlocale(locale.LC_TIME, orig_locale)
    return date



# -*- coding: utf-8 -*-

import unittest
from parameterized import parameterized
import scrapy
import six
from scrapy_rss import FeedItem, RssItem, RssedItem
from tests.utils import RssTestCase


class TestFeedItem(RssTestCase):
    class MyItem1(FeedItem):
        def __init__(self, **kwargs):
            super(self.__class__, self).__init__(**kwargs)
            self.rss = RssItem()

    class MyItem2(FeedItem):
        field = scrapy.Field()
        field2 = scrapy.Field()

        def __init__(self, **kwargs):
            super(self.__class__, self).__init__(**kwargs)
            self.rss = RssItem()

    class MyItem3(RssedItem):
        pass

    class MyItem4(RssedItem):
        field = scrapy.Field()
        field2 = scrapy.Field()

    def test_extendable_item(self):
        from scrapy_rss import ExtendableItem as ExtendableItem1
        from scrapy_rss.meta import ExtendableItem as ExtendableItem2
        from scrapy_rss.meta.item import ExtendableItem as ExtendableItem3

        self.assertIs(ExtendableItem1, FeedItem)
        self.assertIs(ExtendableItem2, FeedItem)
        self.assertIs(ExtendableItem3, FeedItem)

    def test_field_init(self):
        data = {'field': 'value1', 'field2': 2}
        for item_cls in (self.MyItem2, self.MyItem4):
            item = item_cls(**data)
            for key, value in data.items():
                self.assertEqual(item[key], value)

    def test_dict_init(self):
        d = {'field': 'value1', 'field2': 2}
        item = self.MyItem4(d)
        for key, value in d.items():
            self.assertEqual(item[key], value)

    @parameterized.expand([
        ('bad_key', 1),
        ('elements', 'nothing')
    ])
    def test_bad_dict_init(self, key, value):
        for item_cls in (self.MyItem1, self.MyItem2, self.MyItem3, self.MyItem4):
            with six.assertRaisesRegex(self, KeyError, r'does not support field:'):
                item_cls(**{key: value})

    def test_field_setter(self):
        for item in (self.MyItem2(), self.MyItem4()):
            item['field'] = 'value'
            item.rss = None
            item.new_attr = 'OK'

    def test_bad_field_setter(self):
        for item in (self.MyItem2(), self.MyItem4()):
            with six.assertRaisesRegex(self, AttributeError,  r'Use item\[[^\]]+\] = .*? to set field value',
                                       msg="Allowed assignment to Scrapy fields [{} class]"
                                           .format(item.__class__.__name__)):
                item.field = None
            with six.assertRaisesRegex(self, KeyError, r'does not support field:'):
                item['unknown_field'] = 'Bad'

    def test_field_getter(self):
        for item in (self.MyItem2(), self.MyItem4()):
            with six.assertRaisesRegex(self, AttributeError, r'Use item\[[^\]]+\] to get field value'):
                _ = item.field
            _ = item.rss

            item.new_attr = None
            _ = item.new_attr

    def test_uniqueness(self):
        for item_cls in (self.MyItem1, self.MyItem2, self.MyItem3, self.MyItem4):
            item11 = item_cls()
            item12 = item_cls()
            self.assertNotEqual(id(item11.rss), id(item12.rss),
                                msg='[{} class]'.format(item_cls.__name__))

    def test_inheritance(self):
        class Derived1(FeedItem):
            pass

        Derived1()

        class Derived2(FeedItem):
            def __init__(self, **kwargs):
                super(Derived2, self).__init__(**kwargs)

        Derived2()

        class Derived3(RssedItem):
            pass

        Derived3()

        class Derived4(RssedItem):
            def __init__(self, **kwargs):
                super(Derived4, self).__init__(**kwargs)

        Derived4()


if __name__ == '__main__':
    unittest.main()



from os.path import dirname, join
import sys
from setuptools import setup, find_packages


try:
    from wheel.bdist_wheel import bdist_wheel as _bdist_wheel
    import sys

    class bdist_wheel(_bdist_wheel):
        def finalize_options(self):
            _bdist_wheel.finalize_options(self)
            self.python_tag = 'py%i%i' % (sys.version_info.major, sys.version_info.minor)

except ImportError:
    bdist_wheel = None


with open(join(dirname(__file__), 'scrapy_rss/VERSION'), 'rt') as f:
    version = f.read().strip()


with open(join(dirname(__file__), 'requirements.txt'), 'rt') as f:
    install_requires = list(map(str.strip, f))

with open(join(dirname(__file__), 'tests', 'requirements.txt'), 'rt') as f:
    dev_requires = list(map(str.strip, f))

with open('README.rst') as readme:
    setup(
        name='scrapy-rss',
        version=version,
        url='https://github.com/woxcab/scrapy_rss',
        description='RSS Tools for Scrapy Framework',
        long_description=readme.read(),
        license='BSD',
        packages=find_packages(exclude=('tests',)),
        include_package_data=True,
        zip_safe=False,
        python_requires='>=2.7, !=3.0.*, !=3.1.*, !=3.2.*',
        classifiers=[
            'Framework :: Scrapy',
            'Development Status :: 4 - Beta',
            'Intended Audience :: Developers',
            'License :: OSI Approved :: BSD License',
            'Operating System :: OS Independent',
            'Programming Language :: Python',
            'Programming Language :: Python :: 2',
            'Programming Language :: Python :: 2.7',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.3',
            'Programming Language :: Python :: 3.4',
            'Programming Language :: Python :: 3.5',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
            'Programming Language :: Python :: 3.8',
            'Programming Language :: Python :: 3.9',
            'Programming Language :: Python :: 3.10',
            'Topic :: Internet :: WWW/HTTP',
            'Topic :: Software Development :: Libraries :: Python Modules',
        ],
        install_requires=install_requires,
        extras_require={
            'testing': dev_requires
        },
        cmdclass={'bdist_wheel': bdist_wheel},
    )



# -*- coding: utf-8 -*-

import unittest


class TestImport(unittest.TestCase):
    def test_module_level(self):
        from scrapy_rss import RssItem
        from scrapy_rss import FeedItem
        from scrapy_rss import ExtendableItem
        from scrapy_rss import RssedItem
        import scrapy_rss

    def test_elements(self):
        from scrapy_rss.elements import TitleElement
        from scrapy_rss.elements import LinkElement
        from scrapy_rss.elements import DescriptionElement
        from scrapy_rss.elements import AuthorElement
        from scrapy_rss.elements import CategoryElement
        from scrapy_rss.elements import CommentsElement
        from scrapy_rss.elements import EnclosureElement
        from scrapy_rss.elements import GuidElement
        from scrapy_rss.elements import PubDateElement
        from scrapy_rss.elements import SourceElement

    def test_exporters(self):
        from scrapy_rss.exporters import RssItemExporter

    def test_items(self):
        from scrapy_rss.items import RssItem
        from scrapy_rss.items import FeedItem
        from scrapy_rss.items import ExtendableItem
        from scrapy_rss.items import RssedItem

    def test_meta(self):
        from scrapy_rss.meta import BaseNSComponent
        from scrapy_rss.meta import NSComponentName
        from scrapy_rss.meta import ItemElementAttribute
        from scrapy_rss.meta import ItemElementMeta
        from scrapy_rss.meta import ItemElement
        from scrapy_rss.meta import MultipleElements
        from scrapy_rss.meta import ItemMeta
        from scrapy_rss.meta import FeedItem
        from scrapy_rss.meta import ExtendableItem

    def test_pipelines(self):
        from scrapy_rss.pipelines import RssExportPipeline

    def test_utils(self):
        from scrapy_rss.utils import format_rfc822


if __name__ == '__main__':
    unittest.main()
